"""Exports rules to CDF Data Model Storage (DMS) through cognite-sdk."""

import dataclasses
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Literal, cast, no_type_check

import yaml

from cognite.neat.legacy.rules.models.value_types import ValueTypeMapping

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerApplyList,
    ContainerId,
    ContainerProperty,
    DataModelApply,
    DataModelId,
    DirectRelation,
    DirectRelationReference,
    MappedPropertyApply,
    PropertyType,
    SpaceApply,
    ViewApply,
    ViewId,
)
from cognite.client.data_classes.data_modeling.views import (
    ConnectionDefinitionApply,
    SingleHopConnectionDefinitionApply,
)
from cognite.client.exceptions import CogniteAPIError
from pydantic import BaseModel, ConfigDict, field_validator

from cognite.neat.legacy.rules import exceptions
from cognite.neat.legacy.rules.exporters._base import BaseExporter
from cognite.neat.legacy.rules.exporters._validation import are_entity_names_dms_compliant
from cognite.neat.legacy.rules.models._base import ContainerEntity, EntityTypes, ParentClass
from cognite.neat.legacy.rules.models.rules import Class, Property, Rules
from cognite.neat.utils.utils import generate_exception_report

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


@dataclass
class EmptyPropertyType(PropertyType):
    _type = "No property"


@dataclass
class DMSSchema:
    data_model: DataModelApply
    containers: ContainerApplyList


class DMSExporter(BaseExporter[DMSSchema]):
    """Class for exporting transformation rules object to CDF Data Model Storage (DMS).

    Args:
        rules: Transformation rules object.
        data_model_id: The id of the data model to be created.
        container_policy: How to create/reuse existing containers.
        existing_model: In the case of updating an existing model, this is the existing model.
        report: Report. This is used when the exporter object is created from RawRules

    !!! note "Container policy"
        Here is more information about the different container policies:
        - `create`: assumes no containers exist in CDF, will attempt to create them
        - `reuse`: assumes containers exists in CDF, will attempt to only re-use them
        - `extend`: will re-use existing, extend and/or create only missing containers
        - `optimize`: create containers of size's prescribed by NEAT's optimization algorithm
    """

    def __init__(
        self,
        rules: Rules,
        data_model_id: dm.DataModelId | None = None,
        container_policy: Literal["create", "reuse", "extend", "optimize"] = "create",
        existing_model: dm.DataModel[dm.View] | None = None,
        report: str | None = None,
    ):
        super().__init__(rules, report)
        self.data_model_id = data_model_id
        self.container_policy = container_policy
        if container_policy == "extend" and existing_model is None:
            raise ValueError("Container policy is extend-existing, but no existing model is provided")
        if container_policy != "create":
            raise NotImplementedError("Only one-to-one-view container policy is currently supported")
        self.existing_model = existing_model

    def _export_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in {".yaml", ".yml"}:
            warnings.warn("File extension is not .yaml, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".yaml")
        schema = self.export()
        filepath.write_text(
            yaml.safe_dump(
                {
                    "data_models": [schema.data_model.dump(camel_case=True)],
                    "containers": schema.containers.dump(camel_case=True),
                }
            )
        )

    def export(self) -> DMSSchema:
        model = DMSSchemaComponents.from_rules(self.rules, self.data_model_id)
        return DMSSchema(
            data_model=DataModelApply(
                space=model.space,
                external_id=model.external_id,
                version=model.version,
                description=model.description,
                name=model.name,
                views=list(model.views.values()),
            ),
            containers=ContainerApplyList(model.containers.values()),
        )


class DMSSchemaComponents(BaseModel):
    """
    DMS Schema Components pydantic class used to create space(s), containers, views and data model in CDF.

    This can be used to create a data model in CDF from rules.

    Args:
        space: Name of the space to place the resulting data model.
        external_id: External id of the data model.
        version: Version of the data model.
        description: Description of the data model.
        name: Name of the data model.
        containers: Containers connected to the data model.
        views: Views connected to the data model.

    """

    space: str
    external_id: str
    version: str
    description: str | None = None
    name: str | None = None
    containers: dict[str, ContainerApply]
    views: dict[str, ViewApply]

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    @field_validator("views", mode="after")
    def remove_empty_views(cls, value):
        """This validator removes views that do not have any properties or implements."""
        for view_id, view in list(value.items()):
            if not view.properties and not view.implements:
                del value[view_id]

        return value

    @property
    def spaces(self):
        return set(
            [container.space for container in self.containers.values()]
            + [self.space]
            + [view.space for view in self.views.values()]
        )

    @classmethod
    def from_rules(
        cls, rules: Rules, data_model_id: dm.DataModelId | None = None, set_expected_source: bool = True
    ) -> Self:
        """Generates a DataModel class instance from a Rules instance.

        Args:
            rules: instance of Rules.
            data_model_id: The id of the data model to be created.
            set_expected_source: Whether to set the expected source on views with direct relation properties.

        Returns:
            Instance of DataModel.
        """
        if data_model_id and data_model_id.space:
            # update space in rules to match space
            rules.update_space(data_model_id.space)

        if data_model_id and data_model_id.version:
            # update version in rules to match version
            rules.update_version(data_model_id.version)

        if data_model_id and data_model_id.external_id:
            # update data model name to match external_id
            rules.metadata.suffix = data_model_id.external_id

        names_compliant, name_warnings = are_entity_names_dms_compliant(rules, return_report=True)
        if not names_compliant:
            raise exceptions.EntitiesContainNonDMSCompliantCharacters(report=generate_exception_report(name_warnings))

        if rules.metadata.external_id is None:
            raise exceptions.DataModelIdMissing(prefix=rules.metadata.space)

        return cls(
            space=rules.metadata.space,
            external_id=rules.metadata.external_id,
            version=rules.metadata.version,
            description=rules.metadata.description,
            name=rules.metadata.name,
            containers=cls.containers_from_rules(rules),
            views=cls.views_from_rules(rules),
        )

    @classmethod
    def containers_from_rules(cls, rules: Rules) -> dict[str, ContainerApply]:
        """Create a dictionary of ContainerApply instances from a Rules instance.

        Args:
            rules: instance of Rules.`

        Returns:
            Dictionary of ContainerApply instances.
        """

        containers: dict[str, ContainerApply] = {}
        errors: list = []

        for row, property_ in rules.properties.items():
            if property_.container:
                container_property_id: str = cast(str, property_.container_property)
                container_id = property_.container.id

                api_container = cls.as_api_container(property_.container)
                api_container_property = cls.as_api_container_property(property_)

                # scenario: adding new container to the data model for the first time
                if not isinstance(api_container_property.type, EmptyPropertyType) and container_id not in containers:
                    containers[container_id] = api_container
                    containers[container_id].properties[container_property_id] = api_container_property

                # scenario: adding new property to an existing container
                elif (
                    not isinstance(api_container_property.type, EmptyPropertyType)
                    and container_property_id not in containers[container_id].properties
                ):
                    containers[container_id].properties[container_property_id] = api_container_property

                # scenario: property is redefined checking for potential sub-scenarios
                elif (
                    not isinstance(api_container_property.type, EmptyPropertyType)
                    and container_property_id in containers[container_id].properties
                ):
                    existing_property = containers[container_id].properties[container_property_id]

                    # scenario: property is redefined with a different value type -> raise error
                    if not isinstance(existing_property.type, type(api_container_property.type)):
                        errors.append(
                            exceptions.ContainerPropertyValueTypeRedefinition(
                                container_id=container_id,
                                property_id=container_property_id,
                                current_value_type=str(existing_property.type),
                                redefined_value_type=str(api_container_property.type),
                                loc=f"[Properties/Type/{row}]",
                            )
                        )

                    # scenario: default value is redefined -> set default value to None
                    if (
                        not isinstance(existing_property.type, DirectRelation)
                        and not isinstance(api_container_property.type, DirectRelation)
                        and existing_property.default_value != api_container_property.default_value
                    ):
                        containers[container_id].properties[container_property_id] = dataclasses.replace(
                            existing_property, default_value=None
                        )

                    # scenario: property hold multiple values -> set is_list to True
                    if (
                        not isinstance(existing_property.type, DirectRelation)
                        and not isinstance(api_container_property.type, DirectRelation)
                        and existing_property.type.is_list != api_container_property.type.is_list
                    ):
                        containers[container_id].properties[container_property_id].type.is_list = True

        if errors:
            raise ExceptionGroup("Properties value types have been redefined! This is prohibited! Aborting!", errors)

        return containers

    @classmethod
    def as_api_container_property(cls, property_: Property) -> ContainerProperty:
        is_one_to_many = property_.max_count != 1

        # Literal, i.e. Node attribute
        if property_.property_type is EntityTypes.data_property:
            property_type = cast(ValueTypeMapping, property_.expected_value_type.mapping).dms
            return ContainerProperty(
                type=property_type(is_list=is_one_to_many),
                nullable=property_.min_count == 0,
                default_value=property_.default,
                name=property_.property_name,
                description=property_.description,
            )

        # URIRef, i.e. edge 1-1 , aka direct relation between Nodes
        elif property_.property_type is EntityTypes.object_property and not is_one_to_many:
            return ContainerProperty(
                type=DirectRelation(),
                nullable=True,
                name=property_.property_name,
                description=property_.description,
            )

        # return type=None if property cannot be created
        else:
            return ContainerProperty(type=EmptyPropertyType())

    @classmethod
    def as_api_container(cls, container: ContainerEntity) -> ContainerApply:
        return ContainerApply(
            space=container.space,
            external_id=container.external_id,
            description=container.description,
            name=container.name,
            # properties are added later
            properties={},
        )

    @classmethod
    def views_from_rules(cls, rules: Rules) -> dict[str, ViewApply]:
        """Generates a dictionary of ViewApply instances from a Rules instance.

        Args:
            rule: Instance of Rules.
            space: Name of the space to place the views.

        Returns:
            Dictionary of ViewApply instances.
        """
        views: dict[str, ViewApply] = {
            f"{rules.metadata.space}:{class_.class_id}": cls.as_api_view(
                class_, rules.metadata.space, rules.metadata.version
            )
            for class_ in rules.classes.values()
        }
        errors: list = []

        # Create views from property-class definitions
        for row, property_ in rules.properties.items():
            view_property = cls.as_api_view_property(property_, rules.metadata.space)
            id_ = f"{rules.metadata.space}:{property_.class_id}"

            # scenario: view exist but property does not so it is added
            if view_property and (property_.property_id not in cast(dict, views[id_].properties)):
                cast(dict, views[id_].properties)[property_.property_id] = view_property

            # scenario: view exist, property exists but it is differently defined -> raise error
            # type: ignore
            elif (
                view_property
                and property_.property_id in cast(dict, views[id_].properties)
                and view_property is not cast(dict, views[id_].properties)[property_.property_id]
            ):
                errors.append(
                    exceptions.ViewPropertyRedefinition(
                        view_id=id_,
                        property_id=cast(str, property_.property_id),
                        loc=f"[Properties/Property/{row}]",
                    )
                )

        if errors:
            raise ExceptionGroup("View properties have been redefined! This is prohibited! Aborting!", errors)

        return views

    @classmethod
    def as_api_view(cls, class_: Class, space: str, version: str) -> ViewApply:
        return ViewApply(
            space=space,
            external_id=class_.class_id,
            version=version,
            name=class_.class_name,
            description=class_.description,
            properties={},
            implements=(
                [parent_class.view_id for parent_class in cast(list[ParentClass], class_.parent_class)]
                if class_.parent_class
                else None
            ),
        )

    @classmethod
    def as_api_view_property(
        cls, property_: Property, space: str
    ) -> MappedPropertyApply | ConnectionDefinitionApply | None:
        property_.container = cast(ContainerEntity, property_.container)
        property_.container_property = cast(str, property_.container_property)
        if property_.property_type is EntityTypes.data_property:
            return MappedPropertyApply(
                container=ContainerId(space=property_.container.space, external_id=property_.container.external_id),
                container_property_identifier=property_.container_property,
                name=property_.property_name,
                description=property_.description,
            )

        # edge 1-1 == directRelation
        elif property_.property_type is EntityTypes.object_property and property_.max_count == 1:
            return MappedPropertyApply(
                container=ContainerId(space=property_.container.space, external_id=property_.container.external_id),
                container_property_identifier=property_.container_property,
                name=property_.property_name,
                description=property_.description,
                source=ViewId(
                    space=property_.expected_value_type.space,
                    external_id=property_.expected_value_type.external_id,
                    version=property_.expected_value_type.version,
                ),
            )

        # edge 1-many == EDGE, this does not have container! type is here source ViewId ?!
        elif property_.property_type is EntityTypes.object_property and property_.max_count != 1:
            if property_.container and property_.expected_value_type.space != property_.container.space:
                type_ = DirectRelationReference(
                    space=property_.container.space,
                    external_id=f"{property_.container.suffix}.{property_.container_property}",
                )
            else:
                type_ = DirectRelationReference(
                    space=space, external_id=f"{property_.class_id}.{property_.property_id}"
                )

            return SingleHopConnectionDefinitionApply(
                type=type_,
                # Here we create ViewID to the container that the edge is pointing to.
                source=ViewId(
                    space=property_.expected_value_type.space,
                    external_id=property_.expected_value_type.external_id,
                    version=property_.expected_value_type.version,
                ),
                direction="outwards",
                name=property_.property_name,
                description=property_.description,
            )
        else:
            return None

    def find_existing_spaces(self, client: CogniteClient) -> set[str]:
        """Checks if the spaces exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            External ids of spaces which are part of DMS Schema components that already exist in CDF.
        """

        return set(client.data_modeling.spaces.retrieve(list(self.spaces)).as_ids())

    def find_existing_containers(self, client: CogniteClient) -> set[str]:
        """Checks if the containers exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            External ids of containers which are part of DMS Schema components that already exist in CDF.
        """

        return {
            f"{id_.space}:{id_.external_id}"
            for id_ in client.data_modeling.containers.retrieve(
                [container.as_id() for container in self.containers.values()]
            ).as_ids()
        }

    def find_existing_views(self, client: CogniteClient) -> set[str]:
        """Checks if the views exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            External ids of views which are part of DMS Schema components that already exist in CDF.
        """

        return {
            f"{id_.space}:{id_.external_id}"
            for id_ in client.data_modeling.views.retrieve([view.as_id() for view in self.views.values()]).as_ids()
        }

    def find_existing_data_model(self, client: CogniteClient) -> DataModelId | None:
        """Checks if the data model exists in CDF.

        Args:
            client: Cognite client.

        Returns:
             True if the data model exists, False otherwise.
        """
        if model := client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            cdf_data_model = model.latest_version()
            return cdf_data_model.as_id()
        return None

    # mypy unsatisfied with overload , tried all combination and gave up
    @no_type_check
    def to_cdf(
        self,
        client: CogniteClient,
        components_to_create: set | None = None,
        existing_component_handling: Literal["fail", "skip", "update"] = "fail",
        multi_space_components_create: bool = False,
        return_report: bool = False,
    ) -> None | tuple[dict[str, list], dict[str, list]]:
        """Write the the data model to CDF.

        Args:
            client: Connected Cognite client.
            components_to_create: Which components to create. Takes set
            existing_component_handling: How to handle existing components. Takes Literal["fail", "skip", "update"]
            multi_space_comp_create: Whether to create only components belonging to the data model space,
                                     or also additionally components outside of the data model space. Default is False.

        !!! note "Multi Space DMS Schema Components"
            If multi_space_components_create is set to True, the components will be created
            in all spaces used or defined in `Rules`. If set to False, the
            components will only be created in the space defined in `Rules` metadata.

        !!! note "Component Creation Policy"
            Here is more information about the different component creation policies
            configured through components_to_create argument:

            - `all`: all components of the data model will be created, meaning space, containers, views and data model
            - `data model`: only the data model will be created
            - `view`: only views will be created
            - `container`: only the containers will be created

            for creation of containers beyond the data model space, set `multi_space_components_create` argument
            to True.


        !!! note "Existing Component Handling Policy"
            Here is more information about the different existing component handling policies:

            - `fail`: if any component of the data model (DMS schema) already exists
            - `skip`: skip DMS components that exist
            - `update`: create DMS components that do not exist and update those that do exist

            `update` policy is currently not implemented !

        """

        logs, errors = {}, {}

        components_to_create = components_to_create or {"all"}

        existing_spaces = self.find_existing_spaces(client)
        existing_containers = self.find_existing_containers(client)
        existing_views = self.find_existing_views(client)
        existing_data_model = self.find_existing_data_model(client)

        if (
            existing_spaces or existing_containers or existing_data_model or existing_views
        ) and existing_component_handling == "fail":
            raise exceptions.DataModelOrItsComponentsAlreadyExist(
                existing_spaces,
                existing_data_model,
                existing_containers,
                existing_views,
            )

        if "space" in components_to_create or "all" in components_to_create:
            logs["space"], errors["space"] = self.create_space(
                client,
                existing_spaces,
                existing_component_handling == "update",
                multi_space_components_create,
            )
        if "container" in components_to_create or "all" in components_to_create:
            logs["container"], errors["container"] = self.create_containers(
                client,
                existing_containers,
                existing_component_handling == "update",
                multi_space_components_create,
            )

        if "view" in components_to_create or "all" in components_to_create:
            logs["view"], errors["view"] = self.create_views(
                client, existing_views, existing_component_handling == "update"
            )
        if "data model" in components_to_create or "all" in components_to_create:
            logs["data model"], errors["data model"] = self.create_data_model(
                client, existing_data_model, existing_component_handling == "update"
            )

        if return_report:
            return logs, errors

    def create_space(
        self,
        client: CogniteClient,
        existing_spaces: set,
        update: bool = False,
        multi_space_components_create: bool = False,
    ) -> tuple[list, list]:
        logs, errors = [], []

        spaces_to_create = (
            self.spaces - existing_spaces
            if multi_space_components_create
            else ((self.spaces - existing_spaces) & {self.space})
        )

        spaces_to_update = existing_spaces if multi_space_components_create else (existing_spaces & {self.space})

        if spaces_to_create:
            try:
                _ = client.data_modeling.spaces.apply([SpaceApply(space=space) for space in spaces_to_create])
                logs.append(f"Created space {spaces_to_create}")
            except CogniteAPIError as e:
                errors.append(f"Failed to create space {spaces_to_create}! Reason: {e.message}")

        if update and spaces_to_update:
            try:
                _ = client.data_modeling.spaces.apply([SpaceApply(space=space) for space in spaces_to_update])
                logs.append(f"Updated space {spaces_to_update}")
            except CogniteAPIError as e:
                errors.append(f"Failed to update spaces {spaces_to_update}! Reason: {e.message}")

        return logs, errors

    def create_containers(
        self,
        client: CogniteClient,
        existing_containers: set,
        update: bool = False,
        multi_space_components_create: bool = False,
    ) -> tuple[list, list]:
        logs, errors = [], []

        containers_to_create = (
            set(self.containers.keys()) - existing_containers
            if multi_space_components_create
            else {k for k in (set(self.containers.keys()) - existing_containers) if k.split(":")[0] == self.space}
        )

        containers_to_update = (
            existing_containers
            if multi_space_components_create
            else {k for k in existing_containers if k.split(":")[0] == self.space}
        )

        if containers_to_create:
            try:
                _ = client.data_modeling.containers.apply([self.containers[id_] for id_ in containers_to_create])
                logs.append(f"Created container {containers_to_create}")
            except CogniteAPIError as e:
                errors.append(f"Failed to create containers {containers_to_create}! Reason: {e.message}")

        if update and containers_to_update:
            try:
                _ = client.data_modeling.containers.apply([self.containers[id_] for id_ in containers_to_update])
                logs.append(f"Updated containers {containers_to_update}")
            except CogniteAPIError as e:
                errors.append(f"Failed to update containers {containers_to_update}! Reason: {e.message}")

        return logs, errors

    def create_views(
        self,
        client: CogniteClient,
        existing_views: set,
        update: bool = False,
    ) -> tuple[list, list]:
        logs, errors = [], []

        non_existing_views = set(self.views.keys()) - existing_views
        if non_existing_views:
            try:
                _ = client.data_modeling.views.apply([self.views[id_] for id_ in non_existing_views])
                logs.append(f"Created views {non_existing_views}")
            except CogniteAPIError as e:
                errors.append(f"Failed to update views {non_existing_views}! Reason: {e.message}")

        if update and existing_views:
            try:
                _ = client.data_modeling.views.apply([self.views[id_] for id_ in existing_views])
                logs.append(f"Updated views {existing_views}")
            except CogniteAPIError as e:
                errors.append(f"Failed to update views {existing_views}! Reason: {e.message}")

        return logs, errors

    def create_data_model(
        self,
        client: CogniteClient,
        existing_data_model: DataModelId | None = None,
        update: bool = False,
    ) -> tuple[list, list]:
        logs, errors = [], []

        if not existing_data_model:
            try:
                _ = client.data_modeling.data_models.apply(
                    DataModelApply(
                        name=self.name,
                        description=self.description,
                        space=self.space,
                        external_id=self.external_id,
                        version=self.version,
                        views=[view.as_id() for view in self.views.values()],
                    )
                )
                logs.append(f"Created data model {{{self.space}:{self.external_id}/{self.version}}}")
            except CogniteAPIError as e:
                errors.append(
                    f"Failed to create data model {{{self.space}:{self.external_id}/{self.version}}}!"
                    f" Reason: {e.message}"
                )
        elif update:
            try:
                _ = client.data_modeling.data_models.apply(
                    DataModelApply(
                        name=self.name,
                        description=self.description,
                        space=self.space,
                        external_id=self.external_id,
                        version=self.version,
                        views=[view.as_id() for view in self.views.values()],
                    )
                )
                logs.append("Updated data model " f"{{{self.space}:{self.external_id}/{self.version}}}")
            except CogniteAPIError as e:
                errors.append(
                    "Failed to update data model"
                    f" {{{self.space}:{self.external_id}/{self.version}}}! Reason: {e.message}"
                )

        else:
            logs.append(f"Skipped update of data model {{{self.space}:{self.external_id}/{self.version}}}!")

        return logs, errors

    @no_type_check
    def remove(
        self,
        client: CogniteClient,
        components_to_remove: set | None = None,
        multi_space_components_removal: bool = False,
        return_report: bool = False,
    ) -> None | tuple[dict[str, list], dict[str, list]]:
        """Remove DMS schema components from CDF.

        Args:
            client: Connected Cognite client.
            components_to_remove: Which components to remove. Takes set
            multi_space_components_removal: Whether to remove components in multiple spaces,
                                            or only in the space of the data model. Default is False.

        !!! note "Component Creation Policy"
            Here is more information about the different component creation policies:
            - `all`: all components of the data model will be created, meaning space, containers, views and data model
            - `data model`: only the data model will be created
            - `view`: only the views will be created
            - `container`: only the containers will be created


        !!! note "Multi Space DMS Schema Components"
            If multi_space_components_removal is set to True, the components will be removed
            in all spaces used or defined in `Rules`. If set to False, the
            components will only be removed in the space defined in `Rules` metadata.
        """

        logs, errors = {}, {}

        components_to_remove = components_to_remove or {"all"}

        if "data model" in components_to_remove or "all" in components_to_remove:
            logs["data model"], errors["data model"] = self.remove_data_model(client)
        if "view" in components_to_remove or "all" in components_to_remove:
            logs["view"], errors["view"] = self.remove_views(client)
        if "container" in components_to_remove or "all" in components_to_remove:
            logs["container"], errors["container"] = self.remove_containers(client, multi_space_components_removal)
        if "space" in components_to_remove or "all" in components_to_remove:
            logs["space"], errors["space"] = self.remove_spaces(client, multi_space_components_removal)

        if return_report:
            return logs, errors

    def remove_data_model(self, client: CogniteClient) -> tuple[list, list]:
        logs, errors = [], []

        if client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            try:
                _ = client.data_modeling.data_models.delete((self.space, self.external_id, self.version))
                logs.append(f"Removed data model {{{self.space}:{self.external_id}/{self.version}}}")
            except CogniteAPIError as e:
                errors.append(
                    f"Failed to remove data model "
                    f"{{{self.space}:{self.external_id}/{self.version}}}! Reason: {e.message}"
                )
        else:
            logs.append("No Data Model to remove")

        return logs, errors

    def remove_views(self, client: CogniteClient) -> tuple[list, list]:
        logs, errors = [], []

        if existing_views := self.find_existing_views(client):
            try:
                _ = client.data_modeling.views.delete([self.views[id_].as_id() for id_ in existing_views])
                logs.append(f"Removed views {existing_views}!")
            except CogniteAPIError as e:
                errors.append(f"Failed to remove views {existing_views}! Reason: {e.message}")

        else:
            logs.append("No Views to remove!")

        return logs, errors

    def remove_containers(
        self, client: CogniteClient, multi_space_components_removal: bool = False
    ) -> tuple[list, list]:
        logs, errors = [], []

        existing_containers = self.find_existing_containers(client)
        existing_container_in_rules_space = {k for k in existing_containers if k.split(":")[0] == self.space}

        if existing_containers and multi_space_components_removal:
            try:
                _ = client.data_modeling.containers.delete(
                    [self.containers[id_].as_id() for id_ in existing_containers]
                )
                logs.append(f"Removed containers {existing_containers}!")
            except CogniteAPIError as e:
                errors.append(f"Failed to remove containers {existing_containers}! Reason: {e.message}")

        elif existing_container_in_rules_space and not multi_space_components_removal:
            try:
                _ = client.data_modeling.containers.delete(
                    [self.containers[id_].as_id() for id_ in existing_container_in_rules_space]
                )
                logs.append(f"Removed containers {existing_container_in_rules_space}!")
            except CogniteAPIError as e:
                errors.append(f"Failed to remove containers {existing_container_in_rules_space}! Reason: {e.message}")

        else:
            logs.append("No Containers to remove")

        return logs, errors

    def remove_spaces(self, client: CogniteClient, multi_space_components_removal: bool = False) -> tuple[list, list]:
        logs, errors = [], []
        existing_spaces = self.find_existing_spaces(client)

        if existing_spaces and multi_space_components_removal:
            for space in existing_spaces:
                try:
                    _ = client.data_modeling.spaces.delete(space)
                    logs.append(f"Removed spaces {space}!")
                except CogniteAPIError as e:
                    errors.append(f"Failed to remove {space}! Reason: {e.message}")

        elif self.space in existing_spaces and not multi_space_components_removal:
            try:
                _ = client.data_modeling.spaces.delete(self.space)
                logs.append(f"Removed space {self.space}!")
            except CogniteAPIError as e:
                errors.append(f"Failed to remove space {self.space}! Reason: {e.message}")

        else:
            logs.append("No Spaces to remove")

        return logs, errors



from __future__ import annotations

from datetime import date, datetime
from typing import cast

from cognite.client.data_classes.data_modeling import (
    Boolean,
    Date,
    FileReference,
    Float64,
    Int32,
    Int64,
    Json,
    PropertyType,
    SequenceReference,
    Text,
    TimeSeriesReference,
    Timestamp,
)
from pydantic import BaseModel

from cognite.neat.legacy.rules.models._base import Entity, EntityTypes


class ValueTypeMapping(BaseModel):
    """Mapping between XSD, Python, DMS and Graphql types."""

    xsd: str
    python: type
    dms: type
    graphql: str


class ValueType(Entity):
    """Value type is a data/object type defined as a child of Entity model."""

    mapping: ValueTypeMapping | None = None

    @property
    def python(self) -> type | None:
        """Returns the Python type for a given value type."""
        if self.type_ == EntityTypes.data_value_type:
            return cast(ValueTypeMapping, self.mapping).python
        else:
            return None

    @property
    def xsd(self) -> str | None:
        """Returns the XSD type for a given value type."""
        if self.type_ == EntityTypes.data_value_type:
            return cast(ValueTypeMapping, self.mapping).xsd
        else:
            return None

    @property
    def dms(self) -> type | None:
        """Returns the DMS type for a given value type."""
        if self.type_ == EntityTypes.data_value_type:
            return cast(ValueTypeMapping, self.mapping).dms
        else:
            return None

    @property
    def graphql(self) -> str | None:
        """Returns the Graphql type for a given value type."""
        if self.type_ == EntityTypes.data_value_type:
            return cast(ValueTypeMapping, self.mapping).graphql
        else:
            return None


_DATA_TYPES: list[dict] = [
    {"name": "boolean", "python": bool, "GraphQL": "Boolean", "dms": Boolean},
    {"name": "float", "python": float, "GraphQL": "Float", "dms": Float64},
    {"name": "double", "python": float, "GraphQL": "Float", "dms": Float64},
    {"name": "integer", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "nonPositiveInteger", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "nonNegativeInteger", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "negativeInteger", "python": int, "GraphQL": "Int", "dms": Int32},
    {"name": "long", "python": int, "GraphQL": "Int", "dms": Int64},
    {"name": "string", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "langString", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "anyURI", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "normalizedString", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "token", "python": str, "GraphQL": "String", "dms": Text},
    {"name": "PlainLiteral", "python": str, "GraphQL": "String", "dms": Text},
    # Graphql does not have a datetime/date type this is CDF specific
    {"name": "dateTime", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp},
    {"name": "dateTimeStamp", "python": datetime, "GraphQL": "Timestamp", "dms": Timestamp},
    {"name": "date", "python": date, "GraphQL": "String", "dms": Date},
    # CDF specific types, not in XSD
    {"name": "timeseries", "python": TimeSeriesReference, "GraphQL": "TimeSeries", "dms": TimeSeriesReference},
    {"name": "file", "python": FileReference, "GraphQL": "File", "dms": FileReference},
    {"name": "sequence", "python": SequenceReference, "GraphQL": "Sequence", "dms": SequenceReference},
    {"name": "json", "python": Json, "GraphQL": "Json", "dms": Json},
]

XSD_VALUE_TYPE_MAPPINGS: dict[str, ValueType] = {
    data_type["name"]: ValueType(
        prefix="xsd",
        suffix=cast(str, data_type["name"]),
        name=cast(str, data_type["name"]),
        type_=EntityTypes.data_value_type,
        mapping=ValueTypeMapping(
            xsd=data_type["name"],
            python=data_type["python"],
            dms=data_type["dms"],
            graphql=data_type["GraphQL"],
        ),
    )
    for data_type in _DATA_TYPES
}


DMS_VALUE_TYPE_MAPPINGS: dict[type[PropertyType], ValueType] = {}
for value_type in XSD_VALUE_TYPE_MAPPINGS.values():
    if value_type.dms not in DMS_VALUE_TYPE_MAPPINGS:
        DMS_VALUE_TYPE_MAPPINGS[cast(type[PropertyType], value_type.dms)] = cast(ValueType, value_type)



import time
from pathlib import Path
from typing import ClassVar

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import DataModelId

from cognite.neat.rules import importers
from cognite.neat.rules.issues.formatters import FORMATTER_BY_NAME
from cognite.neat.rules.models import RoleTypes
from cognite.neat.rules.models.entities import DataModelEntity, DMSUnknownEntity
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import MultiRuleData
from cognite.neat.workflows.steps.step_model import Configurable, Step

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()

__all__ = [
    "ExcelToRules",
    "OntologyToRules",
    "DMSToRules",
]


class ExcelToRules(Step):
    """This step import rules from the Excel file and validates it."""

    description = "This step imports rules from an excel file "
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File name",
            value="",
            label="Full file name of the rules file in the rules folder. \
                If not provided, step will attempt to get file name from payload \
                    of 'File Uploader' step (if exist)",
        ),
        Configurable(
            name="Report formatter",
            value=next(iter(FORMATTER_BY_NAME.keys())),
            label="The format of the report for the validation of the rules",
            options=list(FORMATTER_BY_NAME),
        ),
        Configurable(
            name="Role",
            value="infer",
            label="For what role Rules are intended?",
            options=["infer", *RoleTypes.__members__.keys()],
        ),
    ]

    def run(self, flow_message: FlowMessage) -> (FlowMessage, MultiRuleData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        file_name = self.configs.get("File name", None)
        full_path = flow_message.payload.get("full_path", None) if flow_message.payload else None

        if file_name:
            rules_file_path = Path(self.data_store_path) / "rules" / file_name
        elif full_path:
            rules_file_path = full_path
        else:
            error_text = "Expected either 'File name' in the step config or 'File uploader' step uploading Excel Rules."
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        # if role is None, it will be inferred from the rules file
        role = self.configs.get("Role")
        role_enum = None
        if role != "infer" and role is not None:
            role_enum = RoleTypes[role]

        excel_importer = importers.ExcelImporter(rules_file_path)
        rules, issues = excel_importer.to_rules(errors="continue", role=role_enum)

        if rules is None:
            output_dir = self.config.staging_path
            report_writer = FORMATTER_BY_NAME[self.configs["Report formatter"]]()
            report_writer.write_to_file(issues, file_or_dir_path=output_dir)
            report_file = report_writer.default_file_name
            error_text = (
                "<p></p>"
                f'<a href="/data/staging/{report_file}?{time.time()}" '
                f'target="_blank">Failed to validate rules, click here for report</a>'
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        output_text = "Rules validation passed successfully!"

        return FlowMessage(output_text=output_text), MultiRuleData.from_rules(rules)


class OntologyToRules(Step):
    """This step import rules from the ontology file (owl) and validates it."""

    description = "This step imports rules from an ontology file "
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File name",
            value="",
            label="Full file name of the ontology file in the rules folder. \
                If not provided, step will attempt to get file name from payload \
                    of 'File Uploader' step (if exist)",
        ),
        Configurable(
            name="Report formatter",
            value=next(iter(FORMATTER_BY_NAME.keys())),
            label="The format of the report for the validation of the rules",
            options=list(FORMATTER_BY_NAME),
        ),
        Configurable(
            name="Role",
            value="infer",
            label="For what role Rules are intended?",
            options=["infer", *RoleTypes.__members__.keys()],
        ),
        Configurable(
            name="Make compliant",
            value="True",
            label=(
                "Attempt to make the imported Rules compliant, by converting "
                "the information model provided in the ontology to data model."
            ),
            options=["True", "False"],
        ),
    ]

    def run(self, flow_message: FlowMessage) -> (FlowMessage, MultiRuleData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        file_name = self.configs.get("File name", None)
        full_path = flow_message.payload.get("full_path", None) if flow_message.payload else None
        make_compliant = self.configs.get("Make compliant", "True") == "True"

        if file_name:
            rules_file_path = self.config.rules_store_path / file_name
        elif full_path:
            rules_file_path = full_path
        else:
            error_text = "Expected either 'File name' in the step config or 'File uploader' step uploading Excel Rules."
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        # if role is None, it will be inferred from the rules file
        role = self.configs.get("Role")
        role_enum = None
        if role != "infer" and role is not None:
            role_enum = RoleTypes[role]

        ontology_importer = importers.OWLImporter(filepath=rules_file_path, make_compliant=make_compliant)
        rules, issues = ontology_importer.to_rules(errors="continue", role=role_enum)

        if rules is None:
            output_dir = self.config.staging_path
            report_writer = FORMATTER_BY_NAME[self.configs["Report formatter"]]()
            report_writer.write_to_file(issues, file_or_dir_path=output_dir)
            report_file = report_writer.default_file_name
            error_text = (
                "<p></p>"
                f'<a href="/data/staging/{report_file}?{time.time()}" '
                f'target="_blank">Failed to validate rules, click here for report</a>'
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        output_text = "Rules validation passed successfully!"

        return FlowMessage(output_text=output_text), MultiRuleData.from_rules(rules)


class DMSToRules(Step):
    """This step imports rules from CDF Data Model"""

    description = "This step imports rules from CDF Data Model"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Data model id",
            value="",
            label="The ID of the Data Model to import. Written at 'my_space:my_data_model(version=1)'",
            type="string",
            required=True,
        ),
        Configurable(
            name="Reference data model id",
            value=None,
            label="The ID of the Reference Data Model to import. Written at 'my_space:my_data_model(version=1)'. "
            "This is typically an enterprise data model when you want to import a solution model",
            type="string",
        ),
        Configurable(
            name="Report formatter",
            value=next(iter(FORMATTER_BY_NAME.keys())),
            label="The format of the report for the validation of the rules",
            options=list(FORMATTER_BY_NAME),
        ),
        Configurable(
            name="Role",
            value="infer",
            label="For what role Rules are intended?",
            options=["infer", *RoleTypes.__members__.keys()],
        ),
    ]

    def run(self, cdf_client: CogniteClient) -> (FlowMessage, MultiRuleData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        datamodel_id_str = self.configs.get("Data model id")
        if datamodel_id_str is None:
            error_text = "Expected input payload to contain 'Data model id' key."
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        datamodel_entity = DataModelEntity.load(datamodel_id_str)
        if isinstance(datamodel_entity, DMSUnknownEntity):
            error_text = (
                f"Data model id should be in the format 'my_space:my_data_model(version=1)' "
                f"or 'my_space:my_data_model', failed to parse space from {datamodel_id_str}"
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        ref_datamodel_str = self.configs.get("Reference data model id", "")
        ref_model_id: DataModelId | None = None
        if ref_datamodel_str:
            ref_model = DataModelEntity.load(ref_datamodel_str)
            if isinstance(ref_model, DMSUnknownEntity):
                error_text = (
                    f"Reference data model id should be in the format 'my_space:my_data_model(version=1)' "
                    f"or 'my_space:my_data_model', failed to parse space from {ref_datamodel_str}"
                )
                return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
            ref_model_id = ref_model.as_id()

        dms_importer = importers.DMSImporter.from_data_model_id(cdf_client, datamodel_entity.as_id(), ref_model_id)

        # if role is None, it will be inferred from the rules file
        role = self.configs.get("Role")
        role_enum = None
        if role != "infer" and role is not None:
            role_enum = RoleTypes[role]

        rules, issues = dms_importer.to_rules(errors="continue", role=role_enum)

        if rules is None:
            output_dir = self.config.staging_path
            report_writer = FORMATTER_BY_NAME[self.configs["Report formatter"]]()
            report_writer.write_to_file(issues, file_or_dir_path=output_dir)
            report_file = report_writer.default_file_name
            error_text = (
                "<p></p>"
                f'<a href="/data/staging/{report_file}?{time.time()}" '
                f'target="_blank">Failed to validate rules, click here for report</a>'
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        output_text = "Rules import and validation passed successfully!"

        return FlowMessage(output_text=output_text), MultiRuleData.from_rules(rules)



import shutil
from collections.abc import Generator, Iterable, Iterator, Mapping
from typing import Any, cast

import pyoxigraph as ox
from rdflib import Graph
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.plugins.sparql.sparql import Query, Update
from rdflib.query import Result
from rdflib.store import VALID_STORE, Store
from rdflib.term import BNode, Identifier, Literal, Node, URIRef, Variable

__all__ = ["OxigraphStore"]

from typing import TypeAlias

_Triple: TypeAlias = tuple[Node, Node, Node]
_Quad: TypeAlias = tuple[Node, Node, Node, Graph]
_TriplePattern: TypeAlias = tuple[Node | None, Node | None, Node | None]


class OxigraphStore(Store):
    context_aware: bool = True
    formula_aware: bool = False
    transaction_aware: bool = False
    graph_aware: bool = True

    def __init__(
        self, configuration: str | None = None, identifier: Identifier | None = None, *, store: ox.Store | None = None
    ):
        self._store = store
        self._prefix_for_namespace: dict[URIRef, str] = {}
        self._namespace_for_prefix: dict[str, URIRef] = {}
        super().__init__(configuration, identifier)

    def open(self, configuration: str, create: bool = False) -> int | None:
        if self._store is not None:
            raise ValueError("The open function should be called before any RDF operation")
        self._store = ox.Store(configuration)
        return VALID_STORE

    def close(self, commit_pending_transaction: bool = False) -> None:
        del self._store

    def destroy(self, configuration: str) -> None:
        shutil.rmtree(configuration)

    def gc(self) -> None:
        pass

    @property
    def _inner(self) -> ox.Store:
        if self._store is None:
            self._store = ox.Store()
        return self._store

    def add(self, triple: _Triple, context: Graph, quoted: bool = False) -> None:
        if quoted:
            raise ValueError("Oxigraph stores are not formula aware")
        self._inner.add(_to_ox(triple, context))
        super().add(triple, context, quoted)

    def addN(self, quads: Iterable[_Quad]) -> None:
        self._inner.extend([_to_ox(q) for q in quads])
        for quad in quads:
            (s, p, o, g) = quad
            super().add((s, p, o), g)

    def remove(self, triple: _TriplePattern, context: Graph | None = None) -> None:
        for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple, context)):
            self._inner.remove(q)
        super().remove(triple, context)

    def triples(
        self, triple_pattern: _TriplePattern, context: Graph | None = None
    ) -> Iterator[tuple[_Triple, Iterator[Graph | None]]]:
        return (_from_ox(q) for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple_pattern, context)))

    def __len__(self, context: Graph | None = None) -> int:
        if context is None:
            # TODO: very bad
            return len({q.triple for q in self._inner})
        return sum(1 for _ in self._inner.quads_for_pattern(None, None, None, _to_ox(context)))

    def contexts(self, triple: _Triple | None = None) -> Generator[Graph, None, None]:
        if triple is None:
            return (_from_ox(g) for g in self._inner.named_graphs())
        return (_from_ox(q[3]) for q in self._inner.quads_for_pattern(*_to_ox_quad_pattern(triple)))

    def query(
        self,
        query: Query | str,
        initNs: Mapping[str, Any],
        initBindings: Mapping[str, Identifier],
        queryGraph: str,
        **kwargs: Any,
    ) -> "Result":
        if isinstance(queryGraph, Query) or kwargs:
            raise NotImplementedError
        init_ns = dict(self._namespace_for_prefix, **initNs)
        if isinstance(query, Query):
            query = str(query)
        query = "".join(f"PREFIX {prefix}: <{namespace}>\n" for prefix, namespace in init_ns.items()) + query
        if initBindings:
            # Todo Anders: This is likely a bug as .n3 is not valid the Identifier.
            #  There are no tests reaching this code.
            query += "\nVALUES ( {} ) {{ ({}) }}".format(
                " ".join(f"?{k}" for k in initBindings),
                " ".join(v.n3() for v in initBindings.values()),  # type: ignore[attr-defined]
            )
        result = self._inner.query(
            query,
            use_default_graph_as_union=queryGraph == "__UNION__",
            default_graph=_to_ox(queryGraph) if isinstance(queryGraph, Node) else None,
        )
        if isinstance(result, bool):
            out = Result("ASK")
            out.askAnswer = result
        elif isinstance(result, ox.QuerySolutions):
            out = Result("SELECT")
            out.vars = [Variable(v.value) for v in result.variables]
            out.bindings = [
                {v: _from_ox(val) for v, val in zip(out.vars, solution, strict=False)} for solution in result
            ]
        elif isinstance(result, ox.QueryTriples):
            out = Result("CONSTRUCT")
            out.graph = Graph()
            out.graph += (_from_ox(t) for t in result)
        else:
            raise ValueError(f"Unexpected query result: {result}")
        return out

    def update(
        self,
        update: Update | str,
        initNs: Mapping[str, Any],
        initBindings: Mapping[str, Identifier],
        queryGraph: str,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    def commit(self) -> None:
        # TODO: implement
        pass

    def rollback(self) -> None:
        # TODO: implement
        pass

    def add_graph(self, graph: Graph) -> None:
        self._inner.add_graph(_to_ox(graph))

    def remove_graph(self, graph: Graph) -> None:
        self._inner.remove_graph(_to_ox(graph))

    def bind(self, prefix: str, namespace: URIRef, override: bool = True) -> None:
        if not override and (prefix in self._namespace_for_prefix or namespace in self._prefix_for_namespace):
            return  # nothing to do
        self._delete_from_prefix(prefix)
        self._delete_from_namespace(namespace)
        self._namespace_for_prefix[prefix] = namespace
        self._prefix_for_namespace[namespace] = prefix

    def _delete_from_prefix(self, prefix):
        if prefix not in self._namespace_for_prefix:
            return
        namespace = self._namespace_for_prefix[prefix]
        del self._namespace_for_prefix[prefix]
        self._delete_from_namespace(namespace)

    def _delete_from_namespace(self, namespace):
        if namespace not in self._prefix_for_namespace:
            return
        prefix = self._prefix_for_namespace[namespace]
        del self._prefix_for_namespace[namespace]
        self._delete_from_prefix(prefix)

    def prefix(self, namespace: URIRef) -> str | None:
        return self._prefix_for_namespace.get(namespace)

    def namespace(self, prefix: str) -> URIRef | None:
        return self._namespace_for_prefix.get(prefix)

    def namespaces(self) -> Iterator[tuple[str, URIRef]]:
        yield from self._namespace_for_prefix.items()


def _to_ox(term: Node | _Triple | _Quad | Graph, context: Graph | None = None):
    if term is None:
        return None
    elif term == DATASET_DEFAULT_GRAPH_ID:
        return ox.DefaultGraph()
    elif isinstance(term, URIRef):
        return ox.NamedNode(term)
    elif isinstance(term, BNode):
        return ox.BlankNode(term)
    elif isinstance(term, Literal):
        return ox.Literal(term, language=term.language, datatype=ox.NamedNode(term.datatype) if term.datatype else None)
    elif isinstance(term, Graph):
        return _to_ox(term.identifier)
    elif isinstance(term, tuple) and len(term) == 3 and isinstance(context, Graph):
        triple = cast(_Triple, term)
        return ox.Quad(_to_ox(triple[0]), _to_ox(triple[1]), _to_ox(triple[2]), _to_ox(context))
    elif isinstance(term, tuple) and len(term) == 4:
        quad = cast(_Quad, term)
        return ox.Quad(_to_ox(quad[0]), _to_ox(quad[1]), _to_ox(quad[2]), _to_ox(quad[3]))
    raise ValueError(f"Unexpected rdflib term: {term!r}")


def _to_ox_quad_pattern(triple: _TriplePattern, context: Graph | None = None):
    (s, p, o) = triple
    return _to_ox_term_pattern(s), _to_ox_term_pattern(p), _to_ox_term_pattern(o), _to_ox_term_pattern(context)


def _to_ox_term_pattern(term):
    if term is None:
        return None
    if isinstance(term, URIRef):
        return ox.NamedNode(term)
    elif isinstance(term, BNode):
        return ox.BlankNode(term)
    elif isinstance(term, Literal):
        return ox.Literal(term, language=term.language, datatype=ox.NamedNode(term.datatype) if term.datatype else None)
    elif isinstance(term, Graph):
        return _to_ox(term.identifier)
    raise ValueError(f"Unexpected rdflib term: {term!r}")


def _from_ox(term):
    if term is None:
        return None
    if isinstance(term, ox.NamedNode):
        return URIRef(term.value)
    if isinstance(term, ox.BlankNode):
        return BNode(term.value)
    if isinstance(term, ox.Literal):
        if term.language:
            return Literal(term.value, lang=term.language)
        return Literal(term.value, datatype=URIRef(term.datatype.value))
    if isinstance(term, ox.DefaultGraph):
        return None
    if isinstance(term, ox.Triple):
        return _from_ox(term.subject), _from_ox(term.predicate), _from_ox(term.object)
    if isinstance(term, ox.Quad):
        return (_from_ox(term.subject), _from_ox(term.predicate), _from_ox(term.object)), _from_ox(term.graph_name)
    raise ValueError(f"Unexpected Oxigraph term: {term!r}")



import math
import sys
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic.main import IncEx
from rdflib import Namespace

from cognite.neat.constants import PREFIXES
from cognite.neat.rules import exceptions, issues
from cognite.neat.rules.issues.base import MultiValueError
from cognite.neat.rules.models._base import (
    BaseMetadata,
    BaseRules,
    DataModelType,
    ExtensionCategory,
    ExtensionCategoryType,
    MatchType,
    RoleTypes,
    SchemaCompleteness,
    SheetEntity,
    SheetList,
)
from cognite.neat.rules.models._rdfpath import (
    AllReferences,
    Hop,
    RawLookup,
    SingleProperty,
    SPARQLQuery,
    TransformationRuleType,
    Traversal,
    parse_rule,
)
from cognite.neat.rules.models._types import (
    NamespaceType,
    PrefixType,
    PropertyType,
    StrListType,
    VersionType,
)
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.domain import DomainRules
from cognite.neat.rules.models.entities import (
    ClassEntity,
    EntityTypes,
    ParentClassEntity,
    ParentEntityList,
    ReferenceEntity,
    Undefined,
    UnknownEntity,
    URLEntity,
)

if TYPE_CHECKING:
    from cognite.neat.rules.models.dms._rules import DMSRules


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class InformationMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.information_architect
    data_model_type: DataModelType = Field(DataModelType.enterprise, alias="dataModelType")
    schema_: SchemaCompleteness = Field(SchemaCompleteness.partial, alias="schema")
    extension: ExtensionCategoryType | None = ExtensionCategory.addition

    prefix: PrefixType
    namespace: NamespaceType

    name: str = Field(
        alias="title",
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, min_length=1, max_length=1024)
    version: VersionType

    created: datetime = Field(
        description=("Date of the data model creation"),
    )

    updated: datetime = Field(
        description=("Date of the data model update"),
    )
    creator: StrListType = Field(
        description=(
            "List of contributors to the data model creation, "
            "typically information architects are considered as contributors."
        ),
    )
    license: str | None = None
    rights: str | None = None

    @model_validator(mode="after")
    def extension_none_but_schema_extend(self) -> Self:
        if self.extension is None:
            self.extension = ExtensionCategory.addition
            return self
        return self

    @field_validator("schema_", mode="plain")
    def as_enum_schema(cls, value: str) -> SchemaCompleteness:
        return SchemaCompleteness(value)

    @field_validator("extension", mode="plain")
    def as_enum_extension(cls, value: str) -> ExtensionCategory:
        return ExtensionCategory(value)

    @field_validator("data_model_type", mode="plain")
    def as_enum_model_type(cls, value: str) -> DataModelType:
        return DataModelType(value)


class InformationClass(SheetEntity):
    """
    Class is a category of things that share a common set of attributes and relationships.

    Args:
        class_: The class ID of the class.
        description: A description of the class.
        parent: The parent class of the class.
        reference: Reference of the source of the information for given resource
        match_type: The match type of the resource being described and the source entity.
    """

    class_: ClassEntity = Field(alias="Class")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    parent: ParentEntityList | None = Field(alias="Parent Class", default=None)
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    match_type: MatchType | None = Field(alias="Match Type", default=None)
    comment: str | None = Field(alias="Comment", default=None)


class InformationProperty(SheetEntity):
    """
    A property is a characteristic of a class. It is a named attribute of a class that describes a range of values
    or a relationship to another class.

    Args:
        class_: Class ID to which property belongs
        property_: Property ID of the property
        name: Property name.
        value_type: Type of value property will hold (data or link to another class)
        min_count: Minimum count of the property values. Defaults to 0
        max_count: Maximum count of the property values. Defaults to None
        default: Default value of the property
        reference: Reference to the source of the information, HTTP URI
        match_type: The match type of the resource being described and the source entity.
        rule_type: Rule type for the transformation from source to target representation
                   of knowledge graph. Defaults to None (no transformation)
        rule: Actual rule for the transformation from source to target representation of
              knowledge graph. Defaults to None (no transformation)
    """

    class_: ClassEntity = Field(alias="Class")
    property_: PropertyType = Field(alias="Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    value_type: DataType | ClassEntity | UnknownEntity = Field(alias="Value Type", union_mode="left_to_right")
    min_count: int | None = Field(alias="Min Count", default=None)
    max_count: int | float | None = Field(alias="Max Count", default=None)
    default: Any | None = Field(alias="Default", default=None)
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    match_type: MatchType | None = Field(alias="Match Type", default=None)
    rule_type: str | TransformationRuleType | None = Field(alias="Rule Type", default=None)
    rule: str | AllReferences | SingleProperty | Hop | RawLookup | SPARQLQuery | Traversal | None = Field(
        alias="Rule", default=None
    )
    comment: str | None = Field(alias="Comment", default=None)

    @field_serializer("max_count", when_used="json-unless-none")
    def serialize_max_count(self, value: int | float | None) -> int | float | None | str:
        if isinstance(value, float) and math.isinf(value):
            return None
        return value

    @field_validator("max_count", mode="before")
    def parse_max_count(cls, value: int | float | None) -> int | float | None:
        if value is None:
            return float("inf")
        return value

    @model_validator(mode="after")
    def is_valid_rule(self):
        # TODO: Can we skip rule_type and simply try to parse the rule and if it fails, raise an error?
        if self.rule_type:
            self.rule_type = self.rule_type.lower()
            if not self.rule:
                raise exceptions.RuleTypeProvidedButRuleMissing(
                    self.property_, self.class_, self.rule_type
                ).to_pydantic_custom_error()
            self.rule = parse_rule(self.rule, self.rule_type)
        return self

    @model_validator(mode="after")
    def set_default_as_list(self):
        if (
            self.type_ == EntityTypes.data_property
            and self.default
            and self.is_list
            and not isinstance(self.default, list)
        ):
            if isinstance(self.default, str):
                if self.default:
                    self.default = self.default.replace(", ", ",").split(",")
                else:
                    self.default = [self.default]
        return self

    @model_validator(mode="after")
    def set_type_for_default(self):
        if self.type_ == EntityTypes.data_property and self.default:
            default_value = self.default[0] if isinstance(self.default, list) else self.default

            if type(default_value) != self.value_type.python:
                try:
                    if isinstance(self.default, list):
                        updated_list = []
                        for value in self.default:
                            updated_list.append(self.value_type.python(value))
                        self.default = updated_list
                    else:
                        self.default = self.value_type.python(self.default)

                except Exception:
                    exceptions.DefaultValueTypeNotProper(
                        self.property_,
                        type(self.default),
                        self.value_type.python,
                    )
        return self

    @property
    def type_(self) -> EntityTypes:
        """Type of property based on value type. Either data (attribute) or object (edge) property."""
        if isinstance(self.value_type, DataType):
            return EntityTypes.data_property
        elif isinstance(self.value_type, ClassEntity):
            return EntityTypes.object_property
        else:
            return EntityTypes.undefined

    @property
    def is_mandatory(self) -> bool:
        """Returns True if property is mandatory."""
        return self.min_count != 0

    @property
    def is_list(self) -> bool:
        """Returns True if property contains a list of values."""
        return self.max_count != 1


class InformationRules(BaseRules):
    metadata: InformationMetadata = Field(alias="Metadata")
    properties: SheetList[InformationProperty] = Field(alias="Properties")
    classes: SheetList[InformationClass] = Field(alias="Classes")
    prefixes: dict[str, Namespace] = Field(default_factory=lambda: PREFIXES.copy())
    last: "InformationRules | None" = Field(None, alias="Last")
    reference: "InformationRules | None" = Field(None, alias="Reference")

    @field_validator("prefixes", mode="before")
    def parse_str(cls, values: Any) -> Any:
        if isinstance(values, dict):
            return {key: Namespace(value) if isinstance(value, str) else value for key, value in values.items()}
        return values

    @model_validator(mode="after")
    def update_entities_prefix(self) -> Self:
        # update expected_value_types
        for property_ in self.properties:
            if isinstance(property_.value_type, ClassEntity) and property_.value_type.prefix is Undefined:
                property_.value_type.prefix = self.metadata.prefix
            if property_.class_.prefix is Undefined:
                property_.class_.prefix = self.metadata.prefix

        # update parent classes
        for class_ in self.classes:
            if class_.parent:
                for parent in cast(list[ParentClassEntity], class_.parent):
                    if not isinstance(parent.prefix, str):
                        parent.prefix = self.metadata.prefix
            if class_.class_.prefix is Undefined:
                class_.class_.prefix = self.metadata.prefix

        return self

    @model_validator(mode="after")
    def post_validation(self) -> "InformationRules":
        from ._validation import InformationPostValidation

        issue_list = InformationPostValidation(self).validate()
        if issue_list.warnings:
            issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError([error for error in issue_list if isinstance(error, issues.NeatValidationError)])
        return self

    def dump(
        self,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        as_reference: bool = False,
    ) -> dict[str, Any]:
        from ._serializer import _InformationRulesSerializer

        dumped = self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
        prefix = self.metadata.prefix
        serializer = _InformationRulesSerializer(by_alias, prefix)
        cleaned = serializer.clean(dumped, as_reference)
        last = "Last" if by_alias else "last"
        if last_dump := cleaned.get(last):
            cleaned[last] = serializer.clean(last_dump, False)
        reference = "Reference" if by_alias else "reference"
        if self.reference and (ref_dump := cleaned.get(reference)):
            prefix = self.reference.metadata.prefix
            cleaned[reference] = _InformationRulesSerializer(by_alias, prefix).clean(ref_dump, True)
        return cleaned

    def as_domain_rules(self) -> DomainRules:
        from ._converter import _InformationRulesConverter

        return _InformationRulesConverter(self).as_domain_rules()

    def as_dms_architect_rules(self) -> "DMSRules":
        from ._converter import _InformationRulesConverter

        return _InformationRulesConverter(self).as_dms_architect_rules()



import math
from typing import Any, ClassVar

from pydantic import Field, field_serializer, field_validator, model_serializer
from pydantic_core.core_schema import SerializationInfo

from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.entities import ClassEntity, ParentEntityList

from ._base import (
    BaseMetadata,
    BaseRules,
    RoleTypes,
    SheetEntity,
    SheetList,
)
from ._types import PropertyType, StrOrListType


class DomainMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.domain_expert
    creator: StrOrListType


class DomainProperty(SheetEntity):
    class_: ClassEntity = Field(alias="Class")
    property_: PropertyType = Field(alias="Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    value_type: DataType | ClassEntity = Field(alias="Value Type")
    min_count: int | None = Field(alias="Min Count", default=None)
    max_count: int | float | None = Field(alias="Max Count", default=None)

    @field_serializer("max_count", when_used="json-unless-none")
    def serialize_max_count(self, value: int | float | None) -> int | float | None | str:
        if isinstance(value, float) and math.isinf(value):
            return None
        return value

    @field_validator("max_count", mode="before")
    def parse_max_count(cls, value: int | float | None) -> int | float | None:
        if value is None:
            return float("inf")
        return value


class DomainClass(SheetEntity):
    class_: ClassEntity = Field(alias="Class")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(None, alias="Description")
    parent: ParentEntityList | None = Field(alias="Parent Class")


class DomainRules(BaseRules):
    metadata: DomainMetadata = Field(alias="Metadata")
    properties: SheetList[DomainProperty] = Field(alias="Properties")
    classes: SheetList[DomainClass] | None = Field(None, alias="Classes")
    last: "DomainRules | None" = Field(None, alias="Last")
    reference: "DomainRules | None" = Field(None, alias="Reference")

    @model_serializer(mode="plain", when_used="always")
    def domain_rules_serializer(self, info: SerializationInfo) -> dict[str, Any]:
        kwargs = vars(info)
        output: dict[str, Any] = {
            "Metadata" if info.by_alias else "metadata": self.metadata.model_dump(**kwargs),
            "Properties" if info.by_alias else "properties": [prop.model_dump(**kwargs) for prop in self.properties],
        }
        if self.classes or not info.exclude_none:
            output["Classes" if info.by_alias else "classes"] = [
                cls.model_dump(**kwargs) for cls in self.classes or []
            ] or None
        return output



from __future__ import annotations

import itertools
import json
from datetime import datetime, timezone
from pathlib import Path
from types import GenericAlias
from typing import Any, ClassVar, Literal, cast, get_args

from openpyxl import Workbook
from openpyxl.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.worksheet.worksheet import Worksheet

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models import (
    DataModelType,
    ExtensionCategory,
    RoleTypes,
    SchemaCompleteness,
    SheetEntity,
)
from cognite.neat.rules.models.dms import DMSMetadata
from cognite.neat.rules.models.domain import DomainMetadata
from cognite.neat.rules.models.information import InformationMetadata

from ._base import BaseExporter


class ExcelExporter(BaseExporter[Workbook]):
    """Export rules to Excel.

    Args:
        styling: The styling to use for the Excel file. Defaults to "default". See below for details
            on the different styles.
        output_role: The role to use for the exported spreadsheet. If provided, the rules will be converted to
            this role formate before being written to excel. If not provided, the role from the rules will be used.
        dump_as: This determines how the rules are written to the Excel file. An Excel file has up to three sets of
           sheets: user, last, and reference. The user sheets are used for inputting rules from a user. The last sheets
           are used for the last version of the same model as the user, while the reference sheets are used for
           the model the user is building on. The options are:
             * "user": The rules are written to the user sheets. This is used when you want to modify the rules
                directly and potentially change the model. This is useful when you have imported the data model
                from outside CDF and you want to modify it before you write it to CDF.
             * "last": The rules are written to the last sheets. This is used when you want to extend the rules,
               but have validation that you are not breaking the existing model. This is used when you want to
               change a model that has already been published to CDF and that model is in production.
             * "reference": The rules are written to the reference sheets. This is typically used when you want to build
               a new solution on top of an enterprise model.
        new_model_id: The new model ID to use for the exported spreadsheet. This is only applicable if the input
            rules have 'is_reference' set. If provided, the model ID will be used to automatically create the
            new metadata sheet in the Excel file. The model id is expected to be a tuple of (prefix, title)
            (space, external_id) for InformationRules and DMSRules respectively.

    The following styles are available:

    - "none":    No styling is applied.
    - "minimal": Column widths are adjusted to fit the content, and the header row(s) is frozen.
    - "default": Minimal + headers are bold, increased size, and colored.
    - "maximal": Default + alternating row colors in the properties sheet for each class in addition to extra
                 blank rows between classes and borders
    """

    Style = Literal["none", "minimal", "default", "maximal"]
    DumpOptions = Literal["user", "last", "reference"]
    _main_header_by_sheet_name: ClassVar[dict[str, str]] = {
        "Properties": "Definition of Properties per Class",
        "Classes": "Definition of Classes",
        "Views": "Definition of Views",
        "Containers": "Definition of Containers",
    }
    style_options = get_args(Style)
    dump_options = get_args(DumpOptions)

    def __init__(
        self,
        styling: Style = "default",
        output_role: RoleTypes | None = None,
        dump_as: DumpOptions = "user",
        new_model_id: tuple[str, str] | None = None,
    ):
        if styling not in self.style_options:
            raise ValueError(f"Invalid styling: {styling}. Valid options are {self.style_options}")
        if dump_as not in self.dump_options:
            raise ValueError(f"Invalid dump_as: {dump_as}. Valid options are {self.dump_options}")
        self.styling = styling
        self._styling_level = self.style_options.index(styling)
        self.output_role = output_role
        self.new_model_id = new_model_id
        self.dump_as = dump_as

    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        """Exports transformation rules to excel file."""
        data = self.export(rules)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None

    def export(self, rules: Rules) -> Workbook:
        rules = self._convert_to_output_role(rules, self.output_role)
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        dumped_user_rules: dict[str, Any]
        dumped_last_rules: dict[str, Any] | None = None
        dumped_reference_rules: dict[str, Any] | None = None
        if self.dump_as != "user":
            action = {"last": "update", "reference": "create"}[self.dump_as]
            metadata_creator = _MetadataCreator(action, self.new_model_id)  # type: ignore[arg-type]

            dumped_user_rules = {
                "Metadata": metadata_creator.create(rules.metadata),
            }

            if self.dump_as == "last":
                dumped_last_rules = rules.dump(by_alias=True)
                if rules.reference:
                    dumped_reference_rules = rules.reference.dump(by_alias=True, as_reference=True)
            elif self.dump_as == "reference":
                dumped_reference_rules = rules.dump(by_alias=True, as_reference=True)
        else:
            dumped_user_rules = rules.dump(by_alias=True)
            if rules.last:
                dumped_last_rules = rules.last.dump(by_alias=True)
            if rules.reference:
                dumped_reference_rules = rules.reference.dump(by_alias=True, as_reference=True)

        self._write_metadata_sheet(workbook, dumped_user_rules["Metadata"])
        self._write_sheets(workbook, dumped_user_rules, rules)
        if dumped_last_rules:
            self._write_sheets(workbook, dumped_last_rules, rules, sheet_prefix="Last")
            self._write_metadata_sheet(workbook, dumped_last_rules["Metadata"], sheet_prefix="Last")

        if dumped_reference_rules:
            self._write_sheets(workbook, dumped_reference_rules, rules, sheet_prefix="Ref")
            self._write_metadata_sheet(workbook, dumped_reference_rules["Metadata"], sheet_prefix="Ref")

        if self._styling_level > 0:
            self._adjust_column_widths(workbook)

        return workbook

    def _write_sheets(self, workbook: Workbook, dumped_rules: dict[str, Any], rules: Rules, sheet_prefix: str = ""):
        for sheet_name, headers in rules.headers_by_sheet(by_alias=True).items():
            if sheet_name in ("Metadata", "prefixes", "Reference", "Last"):
                continue
            sheet = workbook.create_sheet(f"{sheet_prefix}{sheet_name}")

            main_header = self._main_header_by_sheet_name[sheet_name]
            sheet.append([main_header] + [""] * (len(headers) - 1))
            sheet.merge_cells(start_row=1, start_column=1, end_row=1, end_column=len(headers))
            sheet.append(headers)

            fill_colors = itertools.cycle(["CADCFC", "FFFFFF"])
            fill_color = next(fill_colors)
            last_class: str | None = None
            item: dict[str, Any]
            for item in dumped_rules.get(sheet_name) or []:
                row = list(item.values())
                class_ = row[0]

                is_properties = sheet_name == "Properties"
                is_new_class = class_ != last_class and last_class is not None
                if self._styling_level > 2 and is_new_class and is_properties:
                    sheet.append([""] * len(headers))
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                    fill_color = next(fill_colors)

                sheet.append(row)
                if self._styling_level > 2 and is_properties:
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                last_class = class_

            if self._styling_level > 0:
                # This freezes all rows above the given row
                sheet.freeze_panes = sheet["A3"]

                sheet["A1"].alignment = Alignment(horizontal="center")

            if self._styling_level > 1:
                # Make the header row bold, larger, and colored
                sheet["A1"].font = Font(bold=True, size=20)
                sheet["A1"].fill = PatternFill(fgColor="FFC000", patternType="solid")
                for cell in sheet["2"]:
                    cell.font = Font(bold=True, size=14)

    def _write_metadata_sheet(self, workbook: Workbook, metadata: dict[str, Any], sheet_prefix: str = "") -> None:
        # Excel does not support timezone in datetime strings
        if isinstance(metadata.get("created"), datetime):
            metadata["created"] = metadata["created"].replace(tzinfo=None)
        if isinstance(metadata.get("updated"), datetime):
            metadata["updated"] = metadata["updated"].replace(tzinfo=None)

        metadata_sheet = workbook.create_sheet(f"{sheet_prefix}Metadata")
        for key, value in metadata.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    @classmethod
    def _get_item_class(cls, annotation: GenericAlias) -> type[SheetEntity]:
        if not isinstance(annotation, GenericAlias):
            raise ValueError(f"Expected annotation to be a GenericAlias, but got {type(annotation)}")
        args = get_args(annotation)
        if len(args) != 1:
            raise ValueError(f"Expected annotation to have exactly one argument, but got {len(args)}")
        arg = args[0]
        if not issubclass(arg, SheetEntity):
            raise ValueError(f"Expected annotation to have a BaseModel argument, but got {type(arg)}")
        return arg

    @classmethod
    def _adjust_column_widths(cls, workbook: Workbook) -> None:
        for sheet_ in workbook:
            sheet = cast(Worksheet, sheet_)
            for column_cells in sheet.columns:
                try:
                    max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
                except ValueError:
                    max_length = 0

                selected_column = column_cells[0]
                if isinstance(selected_column, MergedCell):
                    selected_column = column_cells[1]

                current = sheet.column_dimensions[selected_column.column_letter].width or (max_length + 0.5)
                sheet.column_dimensions[selected_column.column_letter].width = max(current, max_length + 0.5)
        return None


class _MetadataCreator:
    creator_name = "<YOUR NAME>"

    def __init__(
        self,
        action: Literal["create", "update"],
        new_model_id: tuple[str, str] | None = None,
    ):
        self.action = action
        self.new_model_id = new_model_id or ("YOUR_PREFIX", "YOUR_TITLE")

    def create(self, metadata: DomainMetadata | InformationMetadata | DMSMetadata) -> dict[str, Any]:
        now = datetime.now(timezone.utc).replace(microsecond=0, tzinfo=None)
        if self.action == "update":
            output = json.loads(metadata.model_dump_json(by_alias=True))
            # This is the same for Information and DMS
            output["updated"] = now.isoformat()
            output["schema"] = SchemaCompleteness.extended.value
            output["extension"] = ExtensionCategory.addition.value
            if value := output.get("creator"):
                output["creator"] = f"{value}, {self.creator_name}"
            else:
                output["creator"] = self.creator_name
            return output

        # Action "create"
        if isinstance(metadata, DomainMetadata):
            output = {field_alias: None for field_alias in metadata.model_dump(by_alias=True).keys()}
            output["role"] = metadata.role.value
            output["creator"] = self.creator_name
            return output

        new_metadata = self._create_new_info(now)
        if isinstance(metadata, DMSMetadata):
            from cognite.neat.rules.models.information._converter import _InformationRulesConverter

            output_metadata: DMSMetadata | InformationMetadata = _InformationRulesConverter._convert_metadata_to_dms(
                new_metadata
            )
        elif isinstance(metadata, InformationMetadata):
            output_metadata = new_metadata
        else:
            raise ValueError(f"Bug in Neat: Unknown metadata type: {type(metadata)}")

        created = json.loads(output_metadata.model_dump_json(by_alias=True))
        created.pop("extension", None)
        return created

    def _create_new_info(self, now: datetime) -> InformationMetadata:
        prefix = self.new_model_id[0]
        return InformationMetadata(
            data_model_type=DataModelType.solution,
            schema_=SchemaCompleteness.complete,
            extension=ExtensionCategory.addition,
            prefix=prefix,
            namespace=f"http://purl.org/neat/{prefix}/",  # type: ignore[arg-type]
            description=None,
            version="1",
            created=now,
            updated=now,
            creator=[self.creator_name],
            name=self.new_model_id[1],
        )



from typing import Any, ClassVar, cast

from cognite.neat.rules.models import DMSRules
from cognite.neat.rules.models.dms import DMSContainer, DMSProperty, DMSView
from cognite.neat.rules.models.entities import ReferenceEntity, ViewEntity


class _DMSRulesSerializer:
    # These are the fields that need to be cleaned from the default space and version
    PROPERTIES_FIELDS: ClassVar[list[str]] = ["class_", "view", "value_type", "container"]
    VIEWS_FIELDS: ClassVar[list[str]] = ["class_", "view", "implements"]
    CONTAINERS_FIELDS: ClassVar[list[str]] = ["class_", "container"]

    def __init__(self, by_alias: bool, default_space: str, default_version: str) -> None:
        self.default_space = f"{default_space}:"
        self.default_version = f"version={default_version}"
        self.default_version_wrapped = f"({self.default_version})"

        self.properties_fields = self.PROPERTIES_FIELDS
        self.views_fields = self.VIEWS_FIELDS
        self.containers_fields = self.CONTAINERS_FIELDS
        self.prop_name = "properties"
        self.view_name = "views"
        self.container_name = "containers"
        self.metadata_name = "metadata"
        self.prop_view = "view"
        self.prop_container = "container"
        self.prop_view_property = "view_property"
        self.prop_value_type = "value_type"
        self.view_view = "view"
        self.view_implements = "implements"
        self.container_container = "container"
        self.container_constraint = "constraint"
        self.reference = "Reference" if by_alias else "reference"

        if by_alias:
            self.properties_fields = [
                DMSProperty.model_fields[field].alias or field for field in self.properties_fields
            ]
            self.views_fields = [DMSView.model_fields[field].alias or field for field in self.views_fields]
            self.containers_fields = [
                DMSContainer.model_fields[field].alias or field for field in self.containers_fields
            ]
            self.prop_view = DMSProperty.model_fields[self.prop_view].alias or self.prop_view
            self.prop_container = DMSProperty.model_fields[self.prop_container].alias or self.prop_container
            self.prop_view_property = DMSProperty.model_fields[self.prop_view_property].alias or self.prop_view_property
            self.prop_value_type = DMSProperty.model_fields[self.prop_value_type].alias or self.prop_value_type
            self.view_view = DMSView.model_fields[self.view_view].alias or self.view_view
            self.view_implements = DMSView.model_fields[self.view_implements].alias or self.view_implements
            self.container_container = (
                DMSContainer.model_fields[self.container_container].alias or self.container_container
            )
            self.container_constraint = (
                DMSContainer.model_fields[self.container_constraint].alias or self.container_constraint
            )
            self.prop_name = DMSRules.model_fields[self.prop_name].alias or self.prop_name
            self.view_name = DMSRules.model_fields[self.view_name].alias or self.view_name
            self.container_name = DMSRules.model_fields[self.container_name].alias or self.container_name
            self.metadata_name = DMSRules.model_fields[self.metadata_name].alias or self.metadata_name

    def clean(self, dumped: dict[str, Any], as_reference: bool) -> dict[str, Any]:
        # Sorting to get a deterministic order
        dumped[self.prop_name] = sorted(
            dumped[self.prop_name]["data"], key=lambda p: (p[self.prop_view], p[self.prop_view_property])
        )
        dumped[self.view_name] = sorted(dumped[self.view_name]["data"], key=lambda v: v[self.view_view])
        if container_data := dumped.get(self.container_name):
            dumped[self.container_name] = sorted(container_data["data"], key=lambda c: c[self.container_container])
        else:
            dumped.pop(self.container_name, None)

        for prop in dumped[self.prop_name]:
            if as_reference:
                view_entity = cast(ViewEntity, ViewEntity.load(prop[self.prop_view]))
                prop[self.reference] = str(
                    ReferenceEntity(
                        prefix=view_entity.prefix,
                        suffix=view_entity.suffix,
                        version=view_entity.version,
                        property=prop[self.prop_view_property],
                    )
                )
            for field_name in self.properties_fields:
                if as_reference and field_name == self.prop_container:
                    # When dumping as reference, the container should keep the default space for easy copying
                    # over to user sheets.
                    continue
                if value := prop.get(field_name):
                    prop[field_name] = value.removeprefix(self.default_space).removesuffix(self.default_version_wrapped)
            # Value type can have a property as well
            prop[self.prop_value_type] = prop[self.prop_value_type].replace(self.default_version, "")

        for view in dumped[self.view_name]:
            if as_reference:
                view[self.reference] = view[self.view_view]
            for field_name in self.views_fields:
                if value := view.get(field_name):
                    view[field_name] = value.removeprefix(self.default_space).removesuffix(self.default_version_wrapped)
            if value := view.get(self.view_implements):
                view[self.view_implements] = ",".join(
                    parent.strip().removeprefix(self.default_space).removesuffix(self.default_version_wrapped)
                    for parent in value.split(",")
                )

        for container in dumped.get(self.container_name, []):
            if as_reference:
                container[self.reference] = container[self.container_container]
            for field_name in self.containers_fields:
                if value := container.get(field_name):
                    container[field_name] = value.removeprefix(self.default_space)

            if value := container.get(self.container_constraint):
                container[self.container_constraint] = ",".join(
                    constraint.strip().removeprefix(self.default_space) for constraint in value.split(",")
                )
        return dumped



from pathlib import Path

import pandas as pd

from ._base import BaseImporter


class XSDImporter(BaseImporter):
    """
    Importer for XSD (XML Schema) files.

    Args:
        xml_directory: Path to directory containing XSD files.
    """

    def __init__(self, xsd_directory: Path):
        self.xsd_directory = xsd_directory

    def to_tables(self) -> dict[str, pd.DataFrame]:
        raise NotImplementedError



import re
from collections import Counter
from collections.abc import Iterable
from typing import cast

from rdflib import Graph, Namespace
from rdflib.term import URIRef

from cognite.neat.constants import PREFIXES
from cognite.neat.legacy.rules.analysis import get_classes_with_properties
from cognite.neat.legacy.rules.models._base import Triple
from cognite.neat.legacy.rules.models.rdfpath import (
    AllProperties,
    AllReferences,
    Hop,
    SingleProperty,
    Step,
    TransformationRuleType,
    Traversal,
    parse_rule,
    parse_traversal,
)
from cognite.neat.legacy.rules.models.rules import Rules
from cognite.neat.utils.utils import remove_namespace


def _generate_prefix_header(prefixes: dict[str, Namespace] = PREFIXES) -> str:
    """Generate prefix header which is added to SPARQL query and allows for shorten query statements

    Parameters
    ----------
    prefixes : dict
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    str
        Prefix header
    """
    return "".join(f"PREFIX {key}:<{value}>\n" for key, value in prefixes.items())


def _get_predicate_id(
    graph: Graph, subject_type_id: str, object_type_id: str, prefixes: dict[str, Namespace] = PREFIXES
) -> URIRef:
    """Returns predicate (aka property) URI (i.e., ID) that connects subject and object

    Parameters
    ----------
    graph : Graph
        Data model graph or data model instance (aka knowledge graph)
    subject_type_id : str
        ID of subject type (aka subject class)
    object_type_id : str
        ID of object type (aka object class)
    prefixes : dict, optional
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    URIRef
        ID of predicate (aka property) connecting subject and object
    """
    query = """

        SELECT ?predicateTypeID
        WHERE {
            ?subjectInstanceID a subjectTypeID .
            ?objectInstanceID a objectTypeID .
            ?subjectInstanceID ?predicateTypeID ?objectInstanceID .
        } LIMIT 1"""

    query = query.replace("insertPrefixes", _generate_prefix_header(prefixes))
    final_query = query.replace("subjectTypeID", subject_type_id).replace("objectTypeID", object_type_id)
    res = list(cast(tuple, graph.query(final_query)))

    if len(res) != 1:
        raise ValueError("Subject and Object must have exactly 1 relation!")

    return res[0][0]


def _get_direct_mapping_triples(subject, predicate) -> list[Triple]:
    return [Triple(subject=subject, predicate=predicate, object="?object")]


def _get_all_references_mapping_triples(object) -> list[Triple]:
    return [Triple(subject="?subject", predicate="a", object=object)]


def _get_entire_object_mapping(subject) -> list[Triple]:
    return [Triple(subject=subject, predicate="*")]


def _get_hop_triples(graph, path: Hop, prefixes) -> list[Triple]:
    triples = [Triple(subject="?subject", predicate="a", object=path.class_.id)]
    previous_step = Step(class_=path.class_, direction="origin")

    # add triples for all steps until destination
    for current_step in path.traversal:
        sub_entity, obj_entity = (
            (current_step, previous_step) if current_step.direction == "source" else (previous_step, current_step)
        )

        predicate = _get_predicate_id(graph, sub_entity.class_.id, obj_entity.class_.id, prefixes)

        # if this is first step after origin
        if previous_step.class_.id == path.class_.id:
            if current_step.direction == "source":
                sub, obj = f"?{sub_entity.class_.suffix}ID", "?subject"
            else:
                sub, obj = "?subject", f"?{obj_entity.class_.suffix}ID"

        # Any other step after hoping from origin to first step
        else:
            sub = f"?{sub_entity.class_.suffix}ID"
            obj = f"?{obj_entity.class_.suffix}ID"

        triples.append(Triple(subject=sub, predicate=predicate, object=obj))
        previous_step = current_step

    if previous_step.property:
        triples.extend(
            [
                Triple(subject=f"?{previous_step.class_.suffix}ID", predicate="a", object=previous_step.class_.id),
                Triple(
                    subject=f"?{previous_step.class_.suffix}ID", predicate=previous_step.property.id, object="?object"
                ),
                Triple(subject="?predicate", predicate="a", object=previous_step.property.id),
            ]
        )
    else:
        if previous_step.direction == "source":
            triples[-1].subject = "?object"
        else:
            triples[-1].object = "?object"
        triples.append(Triple(subject="?object", predicate="a", object=previous_step.class_.id))

    return triples


def _get_path_triples(graph: Graph, traversal: Traversal, prefixes: dict[str, Namespace] = PREFIXES) -> list[Triple]:
    """Creates triples subject-predicate-object from declarative graph traversal path

    Parameters
    ----------
    graph : Graph
        Data model graph or data model instance (aka knowledge graph)
    traversal : Traversal
        Parsed traversal path in declarative form
    prefixes : dict, optional
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    list
        List of triples to be consumed by SPARQL query builder
    """
    match traversal:
        case SingleProperty():
            return _get_direct_mapping_triples(traversal.class_.id, traversal.property.id)
        case AllProperties():
            return _get_entire_object_mapping(traversal.class_.id)
        case AllReferences():
            return _get_all_references_mapping_triples(traversal.class_.id)
        case Hop():
            return _get_hop_triples(graph, traversal, prefixes)
        case _:
            raise ValueError("Incorrectly set traversal path!")


BASIC_SPARQL_QUERY_TEMPLATE = """insertPrefixes

SELECT DISTINCT ?subject ?predicate ?object
    WHERE {
query_insertions
        }"""


REFERENCES_ONLY_SPARQL_QUERY_TEMPLATE = """insertPrefixes

SELECT DISTINCT ?subject ?predicate ?object {
    {SELECT DISTINCT  ?object ?predicate {
        query_insertions
        BIND(dct:identifier AS ?predicate)}}

        BIND(?object AS ?subject)

        }"""


# This template does not work with in-memory graph, so it has been replaced with the one above
REFERENCES_ONLY_SPARQL_QUERY_TEMPLATE_OLD = """insertPrefixes

SELECT DISTINCT ?subject ?predicate ?object
    WHERE {
query_insertions
                {
                BIND(?object AS ?subject)
                BIND(dct:identifier AS ?predicate)
                }
        }"""

SINGLE_PROPERTY_SPARQL_QUERY_TEMPLATE = """insertPrefixes

SELECT DISTINCT ?subject ?predicate ?object
    WHERE {
query_insertions
                {
                BIND(property_insertion AS ?predicate)
                }
        }"""


def _generate_all_properties_query_statement(subject: str, query_template: str = BASIC_SPARQL_QUERY_TEMPLATE) -> str:
    query_insertions = "\n".join([f"\t\t?subject a {subject} .", "\t\t?subject ?predicate ?object ."])

    return query_template.replace("query_insertions", query_insertions)


def _generate_all_references_query_statement(
    object: str, query_template: str = REFERENCES_ONLY_SPARQL_QUERY_TEMPLATE
) -> str:
    query_insertions = "\n".join([f"\t\t?object a {object} ."])

    return query_template.replace("query_insertions", query_insertions)


def _generate_single_property_query_statement(
    subject: str, predicate: str, query_template: str = SINGLE_PROPERTY_SPARQL_QUERY_TEMPLATE
) -> str:
    query_insertions = "\n".join([f"\t\t?subject a {subject} .", f"\t\t?subject {predicate} ?object ."])

    return query_template.replace("query_insertions", query_insertions).replace("property_insertion", predicate)


def _generate_hop_query_statement(triples: list[Triple], query_template: str = BASIC_SPARQL_QUERY_TEMPLATE) -> str:
    terminal_triplet = triples[-1]

    query_insertions = "".join(
        f"             {triple.subject} {triple.predicate} {triple.object} .\n" for triple in triples[:-1]
    )

    # Creating terminal query statement based on whether we are query for specific
    # property of an object (first option) or object ID
    if terminal_triplet.subject == "?predicate":
        query_insertions += f"             BIND({terminal_triplet.object} AS ?predicate)\n"
    else:
        query_insertions += (
            f"             {terminal_triplet.subject} {terminal_triplet.predicate} {terminal_triplet.object} .\n"
        )
        query_insertions += "             BIND(dct:relation AS ?predicate)\n"

    return query_template.replace("query_insertions", query_insertions)


def build_sparql_query(
    graph: Graph,
    traversal_path: str | Traversal,
    prefixes: dict[str, Namespace] = PREFIXES,
    insert_prefixes: bool = False,
) -> str:
    """Builds SPARQL query based on declarative traversal path

    Parameters
    ----------
    graph : Graph
        Data model graph or data model instance (aka knowledge graph)
    traversal_path : str
        String representing graph traversal path in declarative form
    prefixes : dict, optional
        Dict containing prefix - namespace pairs, default PREFIXES

    Returns
    -------
    str
        SPARQL query
    """

    traversal = parse_traversal(traversal_path) if isinstance(traversal_path, str) else traversal_path
    triples = _get_path_triples(graph, traversal, prefixes)

    if isinstance(traversal, AllProperties):
        query = _generate_all_properties_query_statement(cast(str, triples[0].subject))
    elif isinstance(traversal, AllReferences) and isinstance(triples[0].object, str):
        query = _generate_all_references_query_statement(triples[0].object)
    elif isinstance(traversal, SingleProperty):
        query = _generate_single_property_query_statement(
            cast(str, triples[0].subject), cast(str, triples[0].predicate)
        )
    elif isinstance(traversal, Hop):
        query = _generate_hop_query_statement(triples)
    else:
        raise ValueError("Not Supported!")

    # Replacing long URIs with short form using their prefixes
    for prefix, URI in prefixes.items():
        query = query.replace(URI, f"{prefix}:")

    return query.replace("insertPrefixes\n\n", _generate_prefix_header(prefixes) if insert_prefixes else "")


def compress_uri(uri: URIRef, prefixes: dict) -> str:
    """Compresses URI to prefix:entity_id

    Parameters
    ----------
    uri : URIRef
        URI of entity
    prefixes : dict
        Dictionary of prefixes

    Returns
    -------
    str
        Compressed URI or original URI if no prefix is found
    """
    return next(
        (
            f"{prefix}:{uri.replace(namespace, '')}"
            for prefix, namespace in prefixes.items()
            if uri.startswith(namespace)
        ),
        uri,
    )


def _hop2property_path(graph: Graph, hop: Hop, prefixes: dict[str, Namespace]) -> str:
    """Converts hop to property path string

    Parameters
    ----------
    graph : Graph
        Graph containing instances of classes
    hop : Hop
        Hop to convert
    prefixes : dict[str, Namespace]
        Dictionary of prefixes to use for compression and predicate quering

    Returns
    -------
    str
        Property path string for hop traversal (e.g. ^rdf:type/rdfs:subClassOf)
    """

    # setting previous step to origin, as we are starting from there
    previous_step = Step(class_=hop.class_, direction="origin")

    # add triples for all steps until destination
    property_path = ""
    for current_step in hop.traversal:
        sub_entity, obj_entity = (
            (current_step, previous_step) if current_step.direction == "source" else (previous_step, current_step)
        )

        predicate_raw = _get_predicate_id(graph, sub_entity.class_.id, obj_entity.class_.id, prefixes)

        predicate = compress_uri(predicate_raw, prefixes)

        predicate = f"^{predicate}" if current_step.direction == "source" else predicate
        property_path += f"{predicate}/"

        previous_step = current_step

    if previous_step.property:
        return property_path + previous_step.property.id
    else:
        # removing "/" at the end of property path if there is no property at the end
        return property_path[:-1]


def build_construct_query(
    graph: Graph,
    class_: str,
    transformation_rules: Rules,
    properties_optional: bool = True,
    class_instances: list[URIRef] | None = None,
) -> str:
    """Builds CONSTRUCT query for given class and rules and optionally filters by class instances

    Parameters
    ----------
    graph : Graph
        Graph containing instances of classes
    class_ : str
        ID of class for which class_instance we want to query
    transformation_rules : TransformationRules
        Transformation rules to use for query generation
    properties_optional : bool, optional
        Whether to make all properties optional, default True
    class_instances : list[URIRef], optional
        List of class instances to filter by, default None (no filter return all instances)

    Returns
    -------
    str
        CONSTRUCT query

    Notes
    -----
    Construct query is far less unforgiving than SELECT query, in sense that it will not return
    anything if one of the properties that define "shape" of the class instance is missing.
    This is the reason why there is option to make all properties optional, so that query will
    return all instances that have at least one property defined.

    """

    query_template = "CONSTRUCT {graph_template\n}\n\nWHERE {graph_pattern\ninsert_filter}"
    query_template = _add_filter(class_instances, query_template)

    templates, patterns = _to_construct_triples(graph, class_, transformation_rules, properties_optional)

    graph_template = "\n           ".join(_triples2sparql_statement(templates))
    graph_pattern = "\n       ".join(_triples2sparql_statement(patterns))

    return query_template.replace("graph_template", graph_template).replace("graph_pattern", graph_pattern)


def _add_filter(class_instances, query_template):
    if class_instances:
        class_instances_formatted = [f"<{instance}>" for instance in class_instances]
        query_template = query_template.replace(
            "insert_filter", f"\n\nFILTER (?subject IN ({', '.join(class_instances_formatted)}))"
        )
    else:
        query_template = query_template.replace("insert_filter", "")
    return query_template


def _triples2sparql_statement(triples: list[Triple]):
    return [
        (
            f"OPTIONAL {{ {triple.subject} {triple.predicate} {triple.object} . }}"
            if triple.optional
            else f"{triple.subject} {triple.predicate} {triple.object} ."
        )
        for triple in triples
    ]


def _to_construct_triples(
    graph: Graph, class_: str, transformation_rules: Rules, properties_optional: bool = True
) -> tuple[list[Triple], list[Triple]]:
    """Converts class definition to CONSTRUCT triples which are used to generate CONSTRUCT query

    Parameters
    ----------
    graph : Graph
        Graph containing instances of classes
    class_ : str
        ID of class for which class_instance we want to query
    transformation_rules : TransformationRules
        Transformation rules to use for query generation

    Returns
    -------
    tuple[list[Triple],list[Triple]]
        Tuple of triples that define graph template and graph pattern parts of CONSTRUCT query
    """
    # TODO: Add handling of UNIONs in rules

    templates = []
    patterns = []

    class_ids = []
    for property_ in get_classes_with_properties(transformation_rules)[class_]:
        if property_.rule_type != TransformationRuleType.rdfpath or property_.skip_rule:
            continue
        if not isinstance(property_.rule, str):
            raise ValueError("Rule must be string!")
        traversal = parse_rule(property_.rule, property_.rule_type).traversal

        if isinstance(traversal, Traversal):
            class_ids.append(traversal.class_.id)

        graph_template_triple = Triple(
            subject="?subject",
            predicate=f"{transformation_rules.metadata.prefix}:{property_.property_id}",
            object=f'?{re.sub(r"[^_a-zA-Z0-9/_]", "_", str(property_.property_id).lower())}',
            optional=False,
        )
        templates.append(graph_template_triple)

        # AllReferences should not be "optional" since we are creating their values
        # by binding them to certain property
        if isinstance(traversal, AllReferences):
            graph_pattern_triple = Triple(
                subject="BIND(?subject", predicate="AS", object=f"{graph_template_triple.object})", optional=False
            )

        elif isinstance(traversal, SingleProperty):
            graph_pattern_triple = Triple(
                subject=graph_template_triple.subject,
                predicate=traversal.property.id,
                object=graph_template_triple.object,
                optional=True if properties_optional else not property_.is_mandatory,
            )

        elif isinstance(traversal, Hop):
            graph_pattern_triple = Triple(
                subject="?subject",
                predicate=_hop2property_path(graph, traversal, transformation_rules.prefixes),
                object=graph_template_triple.object,
                optional=True if properties_optional else not property_.is_mandatory,
            )
        else:
            continue

        patterns.append(graph_pattern_triple)

    # add first triple for graph pattern stating type of object
    patterns.insert(
        0, Triple(subject="?subject", predicate="a", object=_most_occurring_element(class_ids), optional=False)
    )

    return templates, patterns


def _most_occurring_element(list_of_elements: list):
    counts = Counter(list_of_elements)
    return counts.most_common(1)[0][0]


def triples2dictionary(triples: Iterable[tuple[URIRef, URIRef, str | URIRef]]) -> dict[URIRef, dict[str, list[str]]]:
    """Converts list of triples to dictionary"""
    dictionary: dict[URIRef, dict[str, list[str]]] = {}
    for triple in triples:
        id_: str
        property_: str
        value: str
        uri: URIRef
        id_, property_, value = remove_namespace(*triple)  # type: ignore[misc]
        uri = triple[0]

        if uri not in dictionary:
            dictionary[uri] = {"external_id": [id_]}

        if property_ not in dictionary[uri]:
            dictionary[uri][property_] = [value]
        else:
            dictionary[uri][property_].append(value)
    return dictionary



"""This module performs importing of graph to TransformationRules pydantic class.
In more details, it traverses the graph and abstracts class and properties, basically
generating a list of rules based on which nodes that form the graph are made.
"""

import warnings
from datetime import datetime
from typing import cast

import pandas as pd
from rdflib import Graph, Literal, Namespace, URIRef

from cognite.neat.constants import PREFIXES
from cognite.neat.legacy.rules import exceptions
from cognite.neat.legacy.rules.exporters._rules2rules import to_dms_name
from cognite.neat.legacy.rules.models.tables import Tables
from cognite.neat.utils.utils import get_namespace, remove_namespace, uri_to_short_form

from ._base import BaseImporter


class GraphImporter(BaseImporter):
    """
    Convert RDF graph, containing nodes and edges, to tables/ transformation rules / Excel file.

    Args:
        graph: RDF graph to be imported
        max_number_of_instance: Max number of instances to be analyzed for each class in RDF graph


    !!! Note
        Due to high degree of flexibility of RDF graphs, the RDF graph is not guaranteed to be
        converted to a complete and/or valid `Rules` object. Therefore, it is recommended to
        call method `to_raw_rules` to get the raw rules which one should export to Excel file
        using `exporter.ExcelExporter` and then manually edit the Excel file by checking
        validation report file produced by the exporter.

    """

    def __init__(self, graph: Graph, max_number_of_instance: int = -1):
        self.graph = graph
        self.max_number_of_instance = max_number_of_instance

    def to_tables(self) -> dict[str, pd.DataFrame]:
        data_model, prefixes = _graph_to_data_model_dict(self.graph, self.max_number_of_instance)

        return {
            Tables.metadata: _parse_metadata_df(),
            Tables.classes: _parse_classes_df(data_model, prefixes),
            Tables.properties: _parse_properties_df(data_model, prefixes),
            Tables.prefixes: _parse_prefixes_df(prefixes),
        }


def _create_default_properties_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Property pydantic model
    return {
        "header": (
            "Class",
            "Property",
            "Description",
            "Type",
            "Min Count",
            "Max Count",
            "Rule Type",
            "Rule",
            "Source",
            "Source Entity Name",
            "Match Type",
            "Comment",
        )
    }


def _create_default_classes_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Class pydantic model
    return {"header": ("Class", "Description", "Parent Class", "Source", "Source Entity Name", "Match Type", "Comment")}


def _parse_prefixes_df(prefixes: dict[str, Namespace]) -> pd.DataFrame:
    return pd.DataFrame.from_dict({"Prefix": list(prefixes.keys()), "URI": [str(uri) for uri in prefixes.values()]})


def _parse_metadata_df() -> pd.DataFrame:
    clean_list = {
        "namespace": "http://purl.org/cognite/neat/",
        "prefix": "playground",
        "external_id": "neat",
        "version": "1.0.0",
        "isCurrentVersion": True,
        "created": datetime.utcnow(),
        "updated": datetime.utcnow(),
        "title": "RDF Graph Inferred Data Model",
        "description": "This data model has been inferred with NEAT",
        "creator": "NEAT",
        "contributor": "NEAT",
        "rights": "Unknown rights of usage",
        "license": "Unknown license",
    }
    return pd.DataFrame(list(clean_list.items()), columns=["Key", "Value"])


def _parse_classes_df(data_model: dict, prefixes: dict, parsing_config: dict | None = None) -> pd.DataFrame:
    if parsing_config is None:
        parsing_config = _create_default_classes_parsing_config()

    class_rows = []

    for class_ in data_model:
        sanitized_class = to_dms_name(class_, "class")
        class_rows.append(
            [
                sanitized_class,
                None,
                None,
                str(prefixes[data_model[class_]["uri"].split(":")[0]]) + class_,
                class_,
                "exact",
                "Parsed from RDF graph",
            ]
        )

    return pd.DataFrame(class_rows, columns=parsing_config["header"])


def _parse_properties_df(data_model: dict, prefixes: dict, parsing_config: dict | None = None) -> pd.DataFrame:
    if parsing_config is None:
        parsing_config = _create_default_properties_parsing_config()

    property_rows = []

    for class_ in data_model:
        sanitized_class = to_dms_name(class_, "class")
        for property_ in data_model[class_]["properties"]:
            for type_ in data_model[class_]["properties"][property_]["value_type"]:
                sanitized_property = to_dms_name(property_, "property")
                property_rows.append(
                    [
                        sanitized_class,
                        sanitized_property,
                        None,
                        type_,
                        0,  # setting min count to 0 to be more flexible (all properties are optional)
                        max(data_model[class_]["properties"][property_]["occurrence"]),
                        "rdfpath",
                        f'{data_model[class_]["uri"]}({data_model[class_]["properties"][property_]["uri"]})',
                        str(prefixes[data_model[class_]["properties"][property_]["uri"].split(":")[0]]) + property_,
                        property_,
                        "exact",
                        "Parsed from RDF graph",
                    ]
                )

    return pd.DataFrame(property_rows, columns=parsing_config["header"])


def _graph_to_data_model_dict(graph: Graph, max_number_of_instance: int = -1) -> tuple[dict, dict]:
    """Convert RDF graph to dictionary defining data model and prefixes of the graph

    Args:
        graph: RDF graph to be converted to TransformationRules object
        max_number_of_instance: Max number of instances to be considered for each class

    Returns:
        Tuple of data model and prefixes of the graph
    """
    data_model: dict[str, dict] = {}

    prefixes: dict[str, Namespace] = PREFIXES

    for class_ in _get_class_ids(graph):
        _add_uri_namespace_to_prefixes(class_, prefixes)
        class_name = remove_namespace(class_)

        if class_name in data_model:
            warnings.warn(
                exceptions.GraphClassNameCollision(class_name=class_name).message,
                category=exceptions.GraphClassNameCollision,
                stacklevel=2,
            )
            class_name = f"{class_name}_{len(data_model)+1}"

        data_model[class_name] = {"properties": {}, "uri": uri_to_short_form(class_, prefixes)}

        for instance in _get_class_instance_ids(graph, class_, max_number_of_instance):
            for property_, occurrence, data_type, object_type in _define_instance_properties(graph, instance):
                property_name = remove_namespace(property_)
                _add_uri_namespace_to_prefixes(property_, prefixes)

                type_ = data_type if data_type else object_type

                # this is to skip rdf:type property
                if not type_:
                    continue

                type_name = remove_namespace(type_)
                _add_uri_namespace_to_prefixes(type_, prefixes)

                if property_name not in data_model[class_name]["properties"]:
                    data_model[class_name]["properties"][property_name] = {
                        "occurrence": {occurrence.value},
                        "value_type": {type_name: {"uri": uri_to_short_form(type_, prefixes)}},
                        "uri": uri_to_short_form(property_, prefixes),
                    }

                elif type_name not in data_model[class_name]["properties"][property_name]["value_type"]:
                    data_model[class_name]["properties"][property_name]["value_type"][type_name] = {
                        "uri": uri_to_short_form(type_, prefixes)
                    }
                    warnings.warn(
                        exceptions.GraphClassPropertyMultiValueTypes(
                            class_name=class_name,
                            property_name=property_name,
                            types=list(data_model[class_name]["properties"][property_name]["value_type"].keys()),
                        ).message,
                        category=exceptions.GraphClassPropertyMultiValueTypes,
                        stacklevel=3,
                    )

                elif occurrence.value not in data_model[class_name]["properties"][property_name]["occurrence"]:
                    data_model[class_name]["properties"][property_name]["occurrence"].add(occurrence.value)

                    warnings.warn(
                        exceptions.GraphClassPropertyMultiOccurrence(
                            class_name=class_name, property_name=property_name
                        ).message,
                        category=exceptions.GraphClassPropertyMultiOccurrence,
                        stacklevel=3,
                    )
                else:
                    continue

    return data_model, prefixes


def _add_uri_namespace_to_prefixes(URI: URIRef, prefixes: dict[str, Namespace]):
    """Add URI to prefixes dict if not already present

    Args:
        URI: URI from which namespace is being extracted
        prefixes: Dict of prefixes and namespaces
    """
    if Namespace(get_namespace(URI)) not in prefixes.values():
        prefixes[f"prefix-{len(prefixes)+1}"] = Namespace(get_namespace(URI))


def _get_class_ids(graph: Graph) -> list[URIRef]:
    """Get instances ids for a given class

    Args:
        graph: Graph containing class instances
        class_: Class for which instances are to be found
        namespace: Namespace of given class (to avoid writing long URIs)
        limit: Max number of instances to return, by default -1 meaning all instances

    Returns:
        List of class instance URIs
    """

    query_statement = """SELECT ?class (count(?s) as ?instances )
                                WHERE { ?s a ?class . }
                                group by ?class order by DESC(?instances)"""

    return [cast(tuple[URIRef, int], res)[0] for res in list(graph.query(query_statement))]


def _get_class_instance_ids(graph: Graph, class_id: URIRef, max_number_of_instance: int = -1) -> list[URIRef]:
    """Get instances ids for a given class

    Args:
        graph: Graph containing class instances
        class_id: Class id for which instances are to be found

    Returns:
        List of class instance URIs
    """

    query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .}".replace("class", class_id)
    if max_number_of_instance > 0:
        query_statement += f" LIMIT {max_number_of_instance}"
    return [cast(tuple[URIRef], res)[0] for res in list(graph.query(query_statement))]


def _define_instance_properties(
    graph: Graph, instance_id: URIRef
) -> list[tuple[URIRef, Literal, URIRef | None, None | URIRef]]:
    """Get properties of a given instance

    Args:
        graph: Graph containing class instances
        instance_id: Instance id for which properties are to be found and defined

    Returns:
        List of properties of a given instance
    """
    query_statement = """SELECT ?property (count(?property) as ?occurrence) ?dataType ?objectType
                         WHERE {<instance_id> ?property ?value .
                                BIND(datatype(?value) AS ?dataType)
                                OPTIONAL {?value rdf:type ?objectType .}
                                }
                         GROUP BY ?property ?dataType ?objectType"""

    results = graph.query(query_statement.replace("instance_id", instance_id))

    return [cast(tuple[URIRef, Literal, URIRef | None, None | URIRef], res) for res in list(results)]



from typing import Any, cast

from rdflib.query import Result
from rdflib.term import Node


def rdf_result_to_api_response(result: Result) -> dict[str, Any]:
    response: dict[str, Any] = {"fields": [], "rows": []}
    if result.vars is None:
        result.vars = ["s", "p", "o"]  # type: ignore[list-item]

    for field in result.vars:
        response["fields"].append(field)
    for row in result:
        rrow = {field: cast(tuple[Node, Node, Node], row)[i] for i, field in enumerate(result.vars)}
        response["rows"].append(rrow)
    return response



import warnings
from collections import defaultdict
from collections.abc import Sequence
from typing import Any, cast

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling.containers import BTreeIndex
from cognite.client.data_classes.data_modeling.views import (
    SingleEdgeConnectionApply,
    SingleReverseDirectRelationApply,
    ViewPropertyApply,
)

from cognite.neat.rules import issues
from cognite.neat.rules.models._base import DataModelType, ExtensionCategory, SchemaCompleteness
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.entities import (
    ContainerEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    ReferenceEntity,
    ViewEntity,
    ViewPropertyEntity,
)
from cognite.neat.rules.models.wrapped_entities import DMSFilter, HasDataFilter, NodeTypeFilter
from cognite.neat.utils.cdf_classes import ContainerApplyDict, NodeApplyDict, SpaceApplyDict, ViewApplyDict

from ._rules import DMSMetadata, DMSProperty, DMSRules, DMSView
from ._schema import DMSSchema, PipelineSchema


class _DMSExporter:
    """The DMS Exporter is responsible for exporting the DMSRules to a DMSSchema.

    This kept in this location such that it can be used by the DMSRules to validate the schema.
    (This module cannot have a dependency on the exporter module, as it would create a circular dependency.)

    Args
        include_pipeline (bool): If True, the pipeline will be included with the schema. Pipeline means the
            raw tables and transformations necessary to populate the data model.
        instance_space (str): The space to use for the instance. Defaults to None,`Rules.metadata.space` will be used
    """

    def __init__(
        self,
        rules: DMSRules,
        include_pipeline: bool = False,
        instance_space: str | None = None,
    ):
        self.include_pipeline = include_pipeline
        self.instance_space = instance_space
        self.rules = rules
        self._ref_schema = rules.reference.as_schema() if rules.reference else None
        if self._ref_schema:
            # We skip version as that will always be missing in the reference
            self._ref_views_by_id = {
                dm.ViewId(view.space, view.external_id): view for view in self._ref_schema.views.values()
            }
        else:
            self._ref_views_by_id = {}

        self.is_addition = (
            rules.metadata.schema_ is SchemaCompleteness.extended
            and rules.metadata.extension is ExtensionCategory.addition
        )
        self.is_reshape = (
            rules.metadata.schema_ is SchemaCompleteness.extended
            and rules.metadata.extension is ExtensionCategory.reshape
        )
        self.is_rebuild = (
            rules.metadata.schema_ is SchemaCompleteness.extended
            and rules.metadata.extension is ExtensionCategory.rebuild
        )

    def to_schema(self) -> DMSSchema:
        rules = self.rules
        container_properties_by_id, view_properties_by_id = self._gather_properties(list(self.rules.properties))

        # If we are reshaping or rebuilding, and there are no properties in the current rules, we will
        # include those properties from the last rules.
        if rules.last and (self.is_reshape or self.is_rebuild):
            selected_views = {view.view for view in rules.views}
            selected_properties = [
                prop
                for prop in rules.last.properties
                if prop.view in selected_views and prop.view.as_id() not in view_properties_by_id
            ]
            self._update_with_properties(
                selected_properties, container_properties_by_id, view_properties_by_id, include_new_containers=True
            )

        # We need to include the properties from the last rules as well to create complete containers and view
        # depending on the type of extension.
        if rules.last and self.is_addition:
            selected_properties = [
                prop for prop in rules.last.properties if (prop.view.as_id() in view_properties_by_id)
            ]
            self._update_with_properties(selected_properties, container_properties_by_id, view_properties_by_id)
        elif rules.last and (self.is_reshape or self.is_rebuild):
            selected_properties = [
                prop
                for prop in rules.last.properties
                if prop.container and prop.container.as_id() in container_properties_by_id
            ]
            self._update_with_properties(selected_properties, container_properties_by_id, None)

        containers = self._create_containers(container_properties_by_id)

        views, node_types = self._create_views_with_node_types(view_properties_by_id)

        last_schema: DMSSchema | None = None
        if self.rules.last:
            last_schema = self.rules.last.as_schema()
            # Remove the views that are in the current model, last + current should equal the full model
            # without any duplicates
            last_schema.views = ViewApplyDict(
                {view_id: view for view_id, view in last_schema.views.items() if view_id not in views}
            )
            last_schema.containers = ContainerApplyDict(
                {
                    container_id: container
                    for container_id, container in last_schema.containers.items()
                    if container_id not in containers
                }
            )

        views_not_in_model = {view.view.as_id() for view in rules.views if not view.in_model}
        if rules.last and self.is_addition:
            views_not_in_model.update({view.view.as_id() for view in rules.last.views if not view.in_model})

        data_model = rules.metadata.as_data_model()

        data_model_views = [view_id for view_id in views if view_id not in views_not_in_model]
        if last_schema and self.is_addition:
            data_model_views.extend([view_id for view_id in last_schema.views if view_id not in views_not_in_model])

        # Sorting to ensure deterministic order
        data_model.views = sorted(data_model_views, key=lambda v: v.as_tuple())  # type: ignore[union-attr]

        spaces = self._create_spaces(rules.metadata, containers, views, data_model)

        output = DMSSchema(
            spaces=spaces,
            data_model=data_model,
            views=views,
            containers=containers,
            node_types=node_types,
        )
        if self.include_pipeline:
            return PipelineSchema.from_dms(output, self.instance_space)

        if self._ref_schema:
            output.reference = self._ref_schema
        if last_schema:
            output.last = last_schema
        return output

    def _create_spaces(
        self,
        metadata: DMSMetadata,
        containers: ContainerApplyDict,
        views: ViewApplyDict,
        data_model: dm.DataModelApply,
    ) -> SpaceApplyDict:
        used_spaces = {container.space for container in containers.values()} | {view.space for view in views.values()}
        if len(used_spaces) == 1:
            # We skip the default space and only use this space for the data model
            data_model.space = used_spaces.pop()
            spaces = SpaceApplyDict([dm.SpaceApply(space=data_model.space)])
        else:
            used_spaces.add(metadata.space)
            spaces = SpaceApplyDict([dm.SpaceApply(space=space) for space in used_spaces])
        if self.instance_space and self.instance_space not in spaces:
            spaces[self.instance_space] = dm.SpaceApply(space=self.instance_space, name=self.instance_space)
        return spaces

    def _create_views_with_node_types(
        self,
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]],
    ) -> tuple[ViewApplyDict, NodeApplyDict]:
        input_views = list(self.rules.views)
        if self.rules.last:
            existing = {view.view.as_id() for view in input_views}
            modified_views = [
                v
                for v in self.rules.last.views
                if v.view.as_id() in view_properties_by_id and v.view.as_id() not in existing
            ]
            input_views.extend(modified_views)

        views = ViewApplyDict([dms_view.as_view() for dms_view in input_views])
        dms_view_by_id = {dms_view.view.as_id(): dms_view for dms_view in input_views}

        for view_id, view in views.items():
            view.properties = {}
            if not (view_properties := view_properties_by_id.get(view_id)):
                continue
            for prop in view_properties:
                view_property = self._create_view_property(prop, view_properties_by_id)
                if view_property is not None:
                    view.properties[prop.view_property] = view_property

        data_model_type = self.rules.metadata.data_model_type
        unique_node_types: set[dm.NodeId] = set()
        parent_views = {parent for view in views.values() for parent in view.implements or []}
        for view_id, view in views.items():
            dms_view = dms_view_by_id.get(view_id)
            dms_properties = view_properties_by_id.get(view_id, [])
            view_filter = self._create_view_filter(view, dms_view, data_model_type, dms_properties)

            view.filter = view_filter.as_dms_filter()
            if isinstance(view_filter, NodeTypeFilter):
                unique_node_types.update(view_filter.nodes)
                if view.as_id() in parent_views:
                    warnings.warn(issues.dms.NodeTypeFilterOnParentViewWarning(view.as_id()), stacklevel=2)
            elif isinstance(view_filter, HasDataFilter) and data_model_type == DataModelType.solution:
                if dms_view and isinstance(dms_view.reference, ReferenceEntity):
                    references = {dms_view.reference.as_view_id()}
                elif any(True for prop in dms_properties if isinstance(prop.reference, ReferenceEntity)):
                    references = {
                        prop.reference.as_view_id()
                        for prop in dms_properties
                        if isinstance(prop.reference, ReferenceEntity)
                    }
                else:
                    continue
                warnings.warn(
                    issues.dms.HasDataFilterOnViewWithReferencesWarning(view.as_id(), list(references)), stacklevel=2
                )

        return views, NodeApplyDict(
            [dm.NodeApply(space=node.space, external_id=node.external_id) for node in unique_node_types]
        )

    @classmethod
    def _create_edge_type_from_prop(cls, prop: DMSProperty) -> dm.DirectRelationReference:
        if isinstance(prop.reference, ReferenceEntity):
            ref_view_prop = prop.reference.as_view_property_id()
            return cls._create_edge_type_from_view_id(cast(dm.ViewId, ref_view_prop.source), ref_view_prop.property)
        else:
            return cls._create_edge_type_from_view_id(prop.view.as_id(), prop.view_property)

    @staticmethod
    def _create_edge_type_from_view_id(view_id: dm.ViewId, property_: str) -> dm.DirectRelationReference:
        return dm.DirectRelationReference(
            space=view_id.space,
            # This is the same convention as used when converting GraphQL to DMS
            external_id=f"{view_id.external_id}.{property_}",
        )

    def _create_containers(
        self,
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]],
    ) -> ContainerApplyDict:
        containers = list(self.rules.containers or [])
        if self.rules.last:
            existing = {container.container.as_id() for container in containers}
            modified_containers = [
                c
                for c in self.rules.last.containers or []
                if c.container.as_id() in container_properties_by_id and c.container.as_id() not in existing
            ]
            containers.extend(modified_containers)

        containers = dm.ContainerApplyList([dms_container.as_container() for dms_container in containers])
        container_to_drop = set()
        for container in containers:
            container_id = container.as_id()
            if not (container_properties := container_properties_by_id.get(container_id)):
                warnings.warn(issues.dms.EmptyContainerWarning(container_id=container_id), stacklevel=2)
                container_to_drop.add(container_id)
                continue
            for prop in container_properties:
                if prop.container_property is None:
                    continue
                if isinstance(prop.value_type, DataType):
                    type_cls = prop.value_type.dms
                else:
                    type_cls = dm.DirectRelation

                type_ = type_cls(is_list=prop.is_list or False)
                container.properties[prop.container_property] = dm.ContainerProperty(
                    type=type_,
                    nullable=prop.nullable if prop.nullable is not None else True,
                    default_value=prop.default,
                    name=prop.name,
                    description=prop.description,
                )

            uniqueness_properties: dict[str, set[str]] = defaultdict(set)
            for prop in container_properties:
                if prop.container_property is not None:
                    for constraint in prop.constraint or []:
                        uniqueness_properties[constraint].add(prop.container_property)
            for constraint_name, properties in uniqueness_properties.items():
                container.constraints = container.constraints or {}
                container.constraints[constraint_name] = dm.UniquenessConstraint(properties=list(properties))

            index_properties: dict[str, set[str]] = defaultdict(set)
            for prop in container_properties:
                if prop.container_property is not None:
                    for index in prop.index or []:
                        index_properties[index].add(prop.container_property)
            for index_name, properties in index_properties.items():
                container.indexes = container.indexes or {}
                container.indexes[index_name] = BTreeIndex(properties=list(properties))

        # We might drop containers we convert direct relations of list into multi-edge connections
        # which do not have a container.
        for container in containers:
            if container.constraints:
                container.constraints = {
                    name: const
                    for name, const in container.constraints.items()
                    if not (isinstance(const, dm.RequiresConstraint) and const.require in container_to_drop)
                }
        return ContainerApplyDict([container for container in containers if container.as_id() not in container_to_drop])

    @staticmethod
    def _gather_properties(
        properties: Sequence[DMSProperty],
    ) -> tuple[dict[dm.ContainerId, list[DMSProperty]], dict[dm.ViewId, list[DMSProperty]]]:
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]] = defaultdict(list)
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]] = defaultdict(list)
        for prop in properties:
            view_id = prop.view.as_id()
            view_properties_by_id[view_id].append(prop)

            if prop.container and prop.container_property:
                container_id = prop.container.as_id()
                container_properties_by_id[container_id].append(prop)

        return container_properties_by_id, view_properties_by_id

    @classmethod
    def _update_with_properties(
        cls,
        selected_properties: Sequence[DMSProperty],
        container_properties_by_id: dict[dm.ContainerId, list[DMSProperty]],
        view_properties_by_id: dict[dm.ViewId, list[DMSProperty]] | None,
        include_new_containers: bool = False,
    ) -> None:
        view_properties_by_id = view_properties_by_id or {}
        last_container_properties_by_id, last_view_properties_by_id = cls._gather_properties(selected_properties)

        for container_id, properties in last_container_properties_by_id.items():
            # Only add the container properties that are not already present, and do not overwrite.
            if (container_id in container_properties_by_id) or include_new_containers:
                existing = {prop.container_property for prop in container_properties_by_id.get(container_id, [])}
                container_properties_by_id[container_id].extend(
                    [prop for prop in properties if prop.container_property not in existing]
                )

        if view_properties_by_id:
            for view_id, properties in last_view_properties_by_id.items():
                existing = {prop.view_property for prop in view_properties_by_id[view_id]}
                view_properties_by_id[view_id].extend(
                    [prop for prop in properties if prop.view_property not in existing]
                )

    def _create_view_filter(
        self,
        view: dm.ViewApply,
        dms_view: DMSView | None,
        data_model_type: DataModelType,
        dms_properties: list[DMSProperty],
    ) -> DMSFilter:
        selected_filter_name = (dms_view and dms_view.filter_ and dms_view.filter_.name) or ""

        if dms_view and dms_view.filter_ and not dms_view.filter_.is_empty:
            # Has Explicit Filter, use it
            return dms_view.filter_

        if data_model_type == DataModelType.solution and selected_filter_name in [NodeTypeFilter.name, ""]:
            if (
                dms_view
                and isinstance(dms_view.reference, ReferenceEntity)
                and not dms_properties
                and (ref_view := self._ref_views_by_id.get(dms_view.reference.as_view_id()))
                and ref_view.filter
            ):
                # No new properties, only reference, reuse the reference filter
                return DMSFilter.from_dms_filter(ref_view.filter)
            else:
                referenced_node_ids = {
                    prop.reference.as_node_entity()
                    for prop in dms_properties
                    if isinstance(prop.reference, ReferenceEntity)
                }
                if dms_view and isinstance(dms_view.reference, ReferenceEntity):
                    referenced_node_ids.add(dms_view.reference.as_node_entity())
                if referenced_node_ids:
                    return NodeTypeFilter(inner=list(referenced_node_ids))

        # Enterprise Model or (Solution + HasData)
        ref_containers = view.referenced_containers()
        if not ref_containers or selected_filter_name == HasDataFilter.name:
            # Child filter without container properties
            if selected_filter_name == HasDataFilter.name:
                warnings.warn(issues.dms.HasDataFilterOnNoPropertiesViewWarning(view.as_id()), stacklevel=2)
            return NodeTypeFilter(inner=[DMSNodeEntity(space=view.space, externalId=view.external_id)])
        else:
            # HasData or not provided (this is the default)
            return HasDataFilter(inner=[ContainerEntity.from_id(id_) for id_ in ref_containers])

    @classmethod
    def _create_view_property(
        cls, prop: DMSProperty, view_properties_by_id: dict[dm.ViewId, list[DMSProperty]]
    ) -> ViewPropertyApply | None:
        if prop.container and prop.container_property:
            container_prop_identifier = prop.container_property
            extra_args: dict[str, Any] = {}
            if prop.connection == "direct":
                if isinstance(prop.value_type, ViewEntity):
                    extra_args["source"] = prop.value_type.as_id()
                elif isinstance(prop.value_type, DMSUnknownEntity):
                    extra_args["source"] = None
                else:
                    # Should have been validated.
                    raise ValueError(
                        "If this error occurs it is a bug in NEAT, please report"
                        f"Debug Info, Invalid valueType direct: {prop.model_dump_json()}"
                    )
            elif prop.connection is not None:
                # Should have been validated.
                raise ValueError(
                    "If this error occurs it is a bug in NEAT, please report"
                    f"Debug Info, Invalid connection: {prop.model_dump_json()}"
                )
            return dm.MappedPropertyApply(
                container=prop.container.as_id(),
                container_property_identifier=container_prop_identifier,
                name=prop.name,
                description=prop.description,
                **extra_args,
            )
        elif prop.connection == "edge":
            if isinstance(prop.value_type, ViewEntity):
                source_view_id = prop.value_type.as_id()
            else:
                # Should have been validated.
                raise ValueError(
                    "If this error occurs it is a bug in NEAT, please report"
                    f"Debug Info, Invalid valueType edge: {prop.model_dump_json()}"
                )
            edge_cls: type[dm.EdgeConnectionApply] = dm.MultiEdgeConnectionApply
            # If is_list is not set, we default to a MultiEdgeConnection
            if prop.is_list is False:
                edge_cls = SingleEdgeConnectionApply

            return edge_cls(
                type=cls._create_edge_type_from_prop(prop),
                source=source_view_id,
                direction="outwards",
                name=prop.name,
                description=prop.description,
            )
        elif prop.connection == "reverse":
            reverse_prop_id: str | None = None
            if isinstance(prop.value_type, ViewPropertyEntity):
                source_view_id = prop.value_type.as_view_id()
                reverse_prop_id = prop.value_type.property_
            elif isinstance(prop.value_type, ViewEntity):
                source_view_id = prop.value_type.as_id()
            else:
                # Should have been validated.
                raise ValueError(
                    "If this error occurs it is a bug in NEAT, please report"
                    f"Debug Info, Invalid valueType reverse connection: {prop.model_dump_json()}"
                )
            reverse_prop: DMSProperty | None = None
            if reverse_prop_id is not None:
                reverse_prop = next(
                    (
                        prop
                        for prop in view_properties_by_id.get(source_view_id, [])
                        if prop.property_ == reverse_prop_id
                    ),
                    None,
                )

            if reverse_prop is None:
                warnings.warn(
                    issues.dms.ReverseRelationMissingOtherSideWarning(source_view_id, prop.view_property),
                    stacklevel=2,
                )

            if reverse_prop is None or reverse_prop.connection == "edge":
                inwards_edge_cls = (
                    dm.MultiEdgeConnectionApply if prop.is_list in [True, None] else SingleEdgeConnectionApply
                )
                return inwards_edge_cls(
                    type=cls._create_edge_type_from_prop(reverse_prop or prop),
                    source=source_view_id,
                    name=prop.name,
                    description=prop.description,
                    direction="inwards",
                )
            elif reverse_prop_id and reverse_prop and reverse_prop.connection == "direct":
                reverse_direct_cls = (
                    dm.MultiReverseDirectRelationApply if prop.is_list is True else SingleReverseDirectRelationApply
                )
                return reverse_direct_cls(
                    source=source_view_id,
                    through=dm.PropertyId(source=source_view_id, property=reverse_prop_id),
                    name=prop.name,
                    description=prop.description,
                )
            else:
                return None

        elif prop.view and prop.view_property and prop.connection:
            warnings.warn(
                issues.dms.UnsupportedConnectionWarning(prop.view.as_id(), prop.view_property, prop.connection or ""),
                stacklevel=2,
            )
        return None



import getpass
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Literal, overload

from pydantic import ValidationError
from rdflib import Namespace

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.issues.base import IssueList, NeatValidationError, ValidationWarning
from cognite.neat.rules.models import DMSRules, InformationRules, RoleTypes


class BaseImporter(ABC):
    """
    BaseImporter class which all importers inherit from.
    """

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules: ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList]: ...

    @abstractmethod
    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        """
        Creates `Rules` object from the data for target role.
        """
        ...

    @classmethod
    def _to_output(
        cls,
        rules: Rules,
        issues: IssueList,
        errors: Literal["raise", "continue"] = "continue",
        role: RoleTypes | None = None,
    ) -> tuple[Rules | None, IssueList] | Rules:
        """Converts the rules to the output format."""

        if rules.metadata.role is role or role is None:
            output = rules
        elif isinstance(rules, DMSRules) and role is RoleTypes.information_architect:
            output = rules.as_information_architect_rules()
        elif isinstance(rules, InformationRules) and role is RoleTypes.dms_architect:
            output = rules.as_dms_architect_rules()
        else:
            raise NotImplementedError(f"Role {role} is not supported for {type(rules).__name__} rules")

        if errors == "raise":
            return output
        else:
            return output, issues

    @classmethod
    def _return_or_raise(cls, issue_list: IssueList, errors: Literal["raise", "continue"]) -> tuple[None, IssueList]:
        if errors == "raise":
            raise issue_list.as_errors()
        return None, issue_list

    def _default_metadata(self):
        return {
            "prefix": "neat",
            "schema": "partial",
            "namespace": Namespace("http://purl.org/cognite/neat/"),
            "version": "0.1.0",
            "title": "Neat Imported Data Model",
            "created": datetime.now().replace(microsecond=0).isoformat(),
            "updated": datetime.now().replace(microsecond=0).isoformat(),
            "creator": getpass.getuser(),
            "description": f"Imported using {type(self).__name__}",
        }


class _FutureResult:
    def __init__(self) -> None:
        self._result: Literal["success", "failure", "pending"] = "pending"

    @property
    def result(self) -> Literal["success", "failure", "pending"]:
        return self._result


@contextmanager
def _handle_issues(
    issues: IssueList,
    error_cls: type[NeatValidationError] = NeatValidationError,
    warning_cls: type[ValidationWarning] = ValidationWarning,
    error_args: dict[str, Any] | None = None,
) -> Iterator[_FutureResult]:
    """This is an internal help function to handle issues and warnings.

    Args:
        issues: The issues list to append to.
        error_cls: The class used to convert errors to issues.
        warning_cls:  The class used to convert warnings to issues.

    Returns:
        FutureResult: A future result object that can be used to check the result of the context manager.
    """
    with warnings.catch_warnings(record=True) as warning_logger:
        warnings.simplefilter("always")
        future_result = _FutureResult()
        try:
            yield future_result
        except ValidationError as e:
            issues.extend(error_cls.from_pydantic_errors(e.errors(), **(error_args or {})))
            future_result._result = "failure"
        else:
            future_result._result = "success"
        finally:
            if warning_logger:
                issues.extend([warning_cls.from_warning(warning) for warning in warning_logger])



import json
import re
from abc import ABC, abstractmethod
from collections.abc import Collection
from functools import total_ordering
from typing import Any, ClassVar, TypeVar

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import ContainerId, NodeId
from pydantic import BaseModel, model_serializer, model_validator

from cognite.neat.rules.models.entities import ContainerEntity, DMSNodeEntity, Entity


@total_ordering
class WrappedEntity(BaseModel, ABC):
    name: ClassVar[str]
    _inner_cls: ClassVar[type[Entity]]
    inner: list[Entity] | None

    @classmethod
    def load(cls: "type[T_WrappedEntity]", data: Any) -> "T_WrappedEntity":
        if isinstance(data, cls):
            return data
        return cls.model_validate(data)

    @model_validator(mode="before")
    def _load(cls, data: Any) -> dict:
        if isinstance(data, dict):
            return data
        elif not isinstance(data, str):
            raise ValueError(f"Cannot load {cls.__name__} from {data}")
        elif not data.casefold().startswith(cls.name.casefold()):
            raise ValueError(f"Expected {cls.name} but got {data}")
        result = cls._parse(data)
        return result

    @classmethod
    def _parse(cls, data: str) -> dict:
        if data.casefold() == cls.name.casefold():
            return {"inner": None}

        # raw filter case:
        if cls.__name__ == "RawFilter":
            if match := re.search(r"rawFilter\(([\s\S]*?)\)", data):
                return {"filter": match.group(1), "inner": None}
            else:
                raise ValueError(f"Cannot parse {cls.name} from {data}. Ill formatted raw filter.")

        # nodeType and hasData case:
        elif inner := data[len(cls.name) :].removeprefix("(").removesuffix(")"):
            return {"inner": [cls._inner_cls.load(entry.strip()) for entry in inner.split(",")]}
        else:
            raise ValueError(f"Cannot parse {cls.name} from {data}")

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)

    def __str__(self):
        return self.id

    @property
    def id(self) -> str:
        inner = self.as_tuple()[1:]
        return f"{self.name}({','.join(inner)})"

    @property
    def is_empty(self) -> bool:
        return self.inner is None or (isinstance(self.inner, list) and not self.inner)

    def dump(self) -> str:
        return str(self)

    def as_tuple(self) -> tuple[str, ...]:
        entities: list[str] = [str(inner) for inner in self.inner or []]
        return self.name, *entities

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, WrappedEntity):
            return NotImplemented
        return self.as_tuple() < other.as_tuple()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, WrappedEntity):
            return NotImplemented
        return self.as_tuple() == other.as_tuple()

    def __hash__(self) -> int:
        return hash(str(self))

    def __repr__(self) -> str:
        return self.id


T_WrappedEntity = TypeVar("T_WrappedEntity", bound=WrappedEntity)


class DMSFilter(WrappedEntity):
    @abstractmethod
    def as_dms_filter(self, default: Any | None = None) -> dm.filters.Filter:
        raise NotImplementedError

    @classmethod
    def from_dms_filter(cls, filter: dm.Filter) -> "DMSFilter":
        dumped = filter.dump()
        if (body := dumped.get(dm.filters.Equals._filter_name)) and (value := body.get("value")):
            space = value.get("space")
            external_id = value.get("externalId")
            if space is not None and external_id is not None:
                return NodeTypeFilter(inner=[DMSNodeEntity(space=space, externalId=external_id)])
        elif (body := dumped.get(dm.filters.In._filter_name)) and (values := body.get("values")):
            return NodeTypeFilter(
                inner=[
                    DMSNodeEntity(space=entry["space"], externalId=entry["externalId"])
                    for entry in values
                    if isinstance(entry, dict) and "space" in entry and "externalId" in entry
                ]
            )
        elif body := dumped.get(dm.filters.HasData._filter_name):
            return HasDataFilter(
                inner=[
                    ContainerEntity(space=entry["space"], externalId=entry["externalId"])
                    for entry in body
                    if isinstance(entry, dict) and "space" in entry and "externalId" in entry
                ]
            )

        raise ValueError(f"Cannot convert {filter._filter_name} to {cls.__name__}")


class NodeTypeFilter(DMSFilter):
    name: ClassVar[str] = "nodeType"
    _inner_cls: ClassVar[type[DMSNodeEntity]] = DMSNodeEntity
    inner: list[DMSNodeEntity] | None = None  # type: ignore[assignment]

    @property
    def nodes(self) -> list[NodeId]:
        return [node.as_id() for node in self.inner or []]

    def as_dms_filter(self, default: Collection[NodeId] | None = None) -> dm.Filter:
        if self.inner is not None:
            node_ids = [node.as_id() for node in self.inner]
        elif default is not None:
            node_ids = list(default)
        else:
            raise ValueError("Empty nodeType filter, please provide a default node.")
        if len(node_ids) == 1:
            return dm.filters.Equals(
                ["node", "type"], {"space": node_ids[0].space, "externalId": node_ids[0].external_id}
            )
        else:
            return dm.filters.In(
                ["node", "type"],
                [
                    {"space": node.space, "externalId": node.external_id}
                    for node in sorted(node_ids, key=lambda node: node.as_tuple())
                ],
            )


class HasDataFilter(DMSFilter):
    name: ClassVar[str] = "hasData"
    _inner_cls: ClassVar[type[ContainerEntity]] = ContainerEntity
    inner: list[ContainerEntity] | None = None  # type: ignore[assignment]

    def as_dms_filter(self, default: Collection[ContainerId] | None = None) -> dm.Filter:
        containers: list[ContainerId]
        if self.inner:
            containers = [container.as_id() for container in self.inner]
        elif default:
            containers = list(default)
        else:
            raise ValueError("Empty hasData filter, please provide a default containers.")

        return dm.filters.HasData(
            # Sorting to ensure deterministic order
            containers=sorted(containers, key=lambda container: container.as_tuple())  # type: ignore[union-attr]
        )


class RawFilter(DMSFilter):
    name: ClassVar[str] = "rawFilter"
    filter: str
    inner: None = None  # type: ignore[assignment]

    def as_dms_filter(self) -> dm.Filter:  # type: ignore[override]
        try:
            return dm.Filter.load(json.loads(self.filter))
        except json.JSONDecodeError as e:
            raise ValueError(f"Error loading raw filter: {e}") from e

    @property
    def is_empty(self) -> bool:
        return self.filter is None

    def __repr__(self) -> str:
        return self.filter



import logging
import traceback
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import LabelDefinition, LabelDefinitionList

from cognite.neat.legacy.rules.exporters._core import get_labels
from cognite.neat.legacy.rules.models.rules import Rules


def upload_labels(
    client: CogniteClient,
    rules: Rules,
    data_set_id: int,
    extra_labels: list | None = None,
    stop_on_exception: bool = False,
):
    """Generates labels from transformation rules and upload them to CDF

    Args:
        client : Instance of CogniteClient
        transformation_rules : Instance of TransformationRules which contains the labels to upload
        data_set_id : Id of the dataset to upload the labels to
        extra_labels : Any additional labels not defined in TransformationRules ,
                       by default ["historic", "non-historic"]
        stop_on_exception : If this function fails to stop the process, by default False
    """
    if extra_labels is None:
        extra_labels = []

    try:
        logging.debug("Fetching existing labels from CDF")
        # This is to list all labels, not just the ones in the dataset since labels
        # must have unique external_ids, fixing issue that Daniel encountered
        if retrieved_labels := client.labels.list(limit=-1):
            existing_labels = set(retrieved_labels.to_pandas().external_id)
        else:
            existing_labels = set()
        logging.debug(f"Found {len(existing_labels)} existing labels in CDF")
    except Exception as e:
        logging.debug("Error fetching existing labels from CDF, no ")
        traceback.print_exc()
        if stop_on_exception:
            raise e
        existing_labels = set()

    non_existing_labels = set(list(get_labels(rules)) + extra_labels).difference(existing_labels)

    if non_existing_labels:
        logging.debug(f"Creating total of {len(existing_labels)} new labels in CDF")
        labels = [
            LabelDefinition(external_id=label, name=label, data_set_id=data_set_id) for label in non_existing_labels
        ]
        res = cast(LabelDefinitionList, client.labels.create(labels))
        logging.debug(f"Created total of {len(res)} new labels in CDF")
    else:
        logging.debug("No new labels created in CDF")



from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetUpdate, Relationship, RelationshipUpdate
from cognite.client.data_classes.data_modeling import EdgeApply, NodeApply

from cognite.neat.legacy.graph.stores import NeatGraphStoreBase
from cognite.neat.legacy.rules.exporters._rules2dms import DMSSchemaComponents
from cognite.neat.legacy.rules.models.rules import Rules
from cognite.neat.rules.models import DMSRules, DomainRules, InformationRules
from cognite.neat.workflows.steps.step_model import DataContract


class RulesData(DataContract):
    """
    This represents the TransformationRules object.

    Args:
        rules: The TransformationRules object.
    """

    rules: Rules


class MultiRuleData(DataContract):
    domain: DomainRules | None = None
    information: InformationRules | None = None
    dms: DMSRules | None = None

    @classmethod
    def from_rules(cls, rules: DomainRules | InformationRules | DMSRules):
        if isinstance(rules, DomainRules):
            return cls(domain=rules)
        elif isinstance(rules, InformationRules):
            return cls(information=rules)
        elif isinstance(rules, DMSRules):
            return cls(dms=rules)
        else:
            raise ValueError(f"Unsupported rules type {type(rules)}")


class PathData(DataContract):
    """
    This represents the path to the excel file.

    Args:
        excel_file_path: The path to the excel file.
    """

    excel_file_path: Path


class SourceGraph(DataContract):
    """
    This represents the source graph.

    Args:
        graph: The source graph.
    """

    graph: NeatGraphStoreBase


class SolutionGraph(DataContract):
    """
    This represents the solution graph.

    Args:
        graph (NeatGraphStoreBase): The solution graph.

    """

    graph: NeatGraphStoreBase


class ClientData(DataContract):
    """
    This represents an instantiated CogniteClient object.

    Args:
        client: The CogniteClient object.
    """

    client: CogniteClient


class CategorizedAssets(DataContract):
    """ "
    This represents the categorized assets.

    Args:
        assets: The categorized assets.
    """

    assets: dict[str, list[Asset | AssetUpdate]]


class CategorizedRelationships(DataContract):
    """
    This represents the categorized relationships.

    Args:
        relationships: The categorized relationships.

    """

    relationships: dict[str, list[Relationship] | list[RelationshipUpdate]]


class Nodes(DataContract):
    """
    This represents nodes.

    Args:
        nodes: list of nodes.
    """

    nodes: list[NodeApply]


class Edges(DataContract):
    """
    This represents edges.

    Args:
        edges: list of edges.
    """

    edges: list[EdgeApply]


class DMSSchemaComponentsData(DataContract):
    """
    This represents DMS Schema Model.

    Args:
        components: DMS Schema Components model.
    """

    components: DMSSchemaComponents
